from typing import Any, Optional

import numpy as np

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import (
    Bool,
    DependentlyModulatedUnsignedList,
    RangedInteger,
    SizedFloatList,
    SizedRangedIntegerList,
    SizedUnsignedIntegerList,
    SizedUnsignedList,
    String,
    Structure,
    UnsignedFloat,
    UnsignedInteger,
    frozen_attributes,
)

from .componentSystem import ComponentSystem

__all__ = [
    "BindingBaseClass",
    "NoBinding",
    "Linear",
    "Langmuir",
    "LangmuirLDF",
    "LangmuirLDFLiquidPhase",
    "BiLangmuir",
    "BiLangmuirLDF",
    "FreundlichLDF",
    "StericMassAction",
    "AntiLangmuir",
    "Spreading",
    "MobilePhaseModulator",
    "ExtendedMobilePhaseModulator",
    "SelfAssociation",
    "BiStericMassAction",
    "MultistateStericMassAction",
    "SimplifiedMultistateStericMassAction",
    "Saska",
    "GeneralizedIonExchange",
    "HICConstantWaterActivity",
    "HICWaterOnHydrophobicSurfaces",
    "MultiComponentColloidal",
]


@frozen_attributes
class BindingBaseClass(Structure):
    """
    Abstract base class for parameters of binding models.

    Attributes
    ----------
    name : str
        name of the binding model.
    component_system : ComponentSystem
        system of components.
    n_comp : int
        number of components.
    n_binding_sites : int
        Number of binding sites.
        Relevant for Multi-Site isotherms such as Bi-Langmuir.
        The default is 1.
    bound_states : list of unsigned integers.
        Number of binding sites per component.
    non_binding_component_indices : list
        (Hardcoded) list of non binding modifier components (e.g. pH).
    is_kinetic : bool
        If False, adsorption is assumed to be in rapid equilibriu.
        The default is True.
    parameters : dict
        dict with parameter values.
    """

    name = String()
    is_kinetic = Bool(default=True)

    n_binding_sites = RangedInteger(lb=1, ub=1, default=1)
    _bound_states = SizedRangedIntegerList(
        size=("n_binding_sites", "n_comp"), lb=0, ub=1, default=1
    )
    non_binding_component_indices = []

    _parameters = ["is_kinetic"]

    def __init__(
        self,
        component_system: ComponentSystem,
        name: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize binding model.

        Parameters
        ----------
        component_system: ComponentSystem
            Component system of the binding model.
        name: str
            Name of the binding model.

        """
        self.component_system = component_system
        self.name = name

        super().__init__(*args, **kwargs)

    @property
    def model(self) -> str:
        """str: Name of the binding model."""
        return self.__class__.__name__

    @property
    def component_system(self) -> ComponentSystem:
        """ComponentSystem: Component system of the binding model."""
        return self._component_system

    @component_system.setter
    def component_system(self, component_system: ComponentSystem) -> None:
        if not isinstance(component_system, ComponentSystem):
            raise TypeError("Expected ComponentSystem")
        self._component_system = component_system

    @property
    def n_comp(self) -> int:
        """int: Number of components."""
        return self.component_system.n_comp

    @property
    def bound_states(self) -> list[int]:
        """list[int]: Number of bound states per component."""
        bound_states = self._bound_states
        for i in self.non_binding_component_indices:
            bound_states[i] = 0
        return bound_states

    @bound_states.setter
    def bound_states(self, bound_states: np.ndarray) -> None:
        indices = self.non_binding_component_indices
        if any(bound_states[i] > 0 for i in indices):
            raise CADETProcessError("Cannot set bound state for non-binding component.")

        self._bound_states = bound_states

    @property
    def n_bound_states(self) -> int:
        """int: Number of bound states."""
        return sum(self.bound_states)

    def __repr__(self) -> str:
        """str: String representation of the binding model."""
        return f"{self.__class__.__name__}(\
            component_system={self.component_system}, name={self.name})')"

    def __str__(self) -> str:
        """str: String representation of the binding model."""
        if self.name is None:
            return self.__class__.__name__
        return self.name


class NoBinding(BindingBaseClass):
    """
    Dummy class for units that do not experience binging behavior.

    The number of components is set to zero for this class.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize NoBinding."""
        super().__init__(ComponentSystem(), name="NoBinding")


class Linear(BindingBaseClass):
    """
    Linear binding model.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants. Length depends on `n_comp`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants. Length depends on `n_comp`.
    """

    adsorption_rate = SizedUnsignedList(size="n_comp")
    desorption_rate = SizedUnsignedList(size="n_comp")

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
    ]


class Langmuir(BindingBaseClass):
    """
    Multi Component Langmuir binding model.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants. Length depends on `n_comp`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants. Length depends on `n_comp`.
    capacity : list of unsigned floats.
        Maximum adsorption capacities. Length depends on `n_comp`.
    """

    adsorption_rate = SizedUnsignedList(size="n_comp")
    desorption_rate = SizedUnsignedList(size="n_comp")
    capacity = SizedUnsignedList(size="n_comp")

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
        "capacity",
    ]


class LangmuirLDF(BindingBaseClass):
    """
    Multi Component Langmuir binding model with linear driving force approximation.

    Note, this variant is based on the equilibrium concentration q* for given c.

    Attributes
    ----------
    equilibrium_constant : list of unsigned floats.
        Adsorption rate constants. Length depends on `n_comp`.
    driving_force_coefficient : list of unsigned floats.
        Desorption rate constants. Length depends on `n_comp`.
    capacity : list of unsigned floats.
        Maximum adsorption capacities. Length depends on `n_comp`.
    """

    equilibrium_constant = SizedUnsignedList(size="n_comp")
    driving_force_coefficient = SizedUnsignedList(size="n_comp")
    capacity = SizedUnsignedList(size="n_comp")

    _parameters = [
        "equilibrium_constant",
        "driving_force_coefficient",
        "capacity",
    ]


class LangmuirLDFLiquidPhase(BindingBaseClass):
    """
    Multi Component Langmuir binding model with linear driving force approximation.

    Note, this variant is based on the equilibrium concentration c* for given q.

    Attributes
    ----------
    equilibrium_constant : list of unsigned floats.
        Adsorption rate constants. Length depends on `n_comp`.
    driving_force_coefficient : list of unsigned floats.
        Desorption rate constants. Length depends on `n_comp`.
    capacity : list of unsigned floats.
        Maximum adsorption capacities. Length depends on `n_comp`.
    """

    equilibrium_constant = SizedUnsignedList(size="n_comp")
    driving_force_coefficient = SizedUnsignedList(size="n_comp")
    capacity = SizedUnsignedList(size="n_comp")

    _parameters = [
        "equilibrium_constant",
        "driving_force_coefficient",
        "capacity",
    ]


class BiLangmuir(BindingBaseClass):
    """
    Multi Component Bi-Langmuir binding model.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants in state-major ordering.
        Length depends on `n_bound_states`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants in state-major ordering.
        Length depends on `n_bound_states`.
    capacity : list of unsigned floats.
        Maximum adsorption capacities in state-major ordering.
        Length depends on `n_bound_states`.
    """

    n_binding_sites = UnsignedInteger(default=2)

    adsorption_rate = SizedUnsignedList(size="n_bound_states")
    desorption_rate = SizedUnsignedList(size="n_bound_states")
    capacity = SizedUnsignedList(size="n_bound_states")

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
        "capacity",
    ]

    def __init__(self, *args: Any, n_binding_sites: int = 2, **kwargs: Any) -> None:
        """Initialize BiLangmuir."""
        self.n_binding_sites = n_binding_sites

        super().__init__(*args, **kwargs)


class BiLangmuirLDF(BindingBaseClass):
    """
    Multi Component Bi-Langmuir binding model.

    Attributes
    ----------
    equilibrium_constant : list of unsigned floats.
        Adsorption rate constants in state-major ordering.
        Length depends on `n_bound_states`.
    driving_force_coefficient : list of unsigned floats.
        Desorption rate constants in state-major ordering.
        Length depends on `n_bound_states`.
    capacity : list of unsigned floats.
        Maximum adsorption capacities in state-major ordering.
        Length depends on `n_bound_states`.
    """

    n_binding_sites = UnsignedInteger(default=2)

    equilibrium_constant = SizedUnsignedList(size="n_bound_states")
    driving_force_coefficient = SizedUnsignedList(size="n_bound_states")
    capacity = SizedUnsignedList(size="n_bound_states")

    _parameters = [
        "equilibrium_constant",
        "driving_force_coefficient",
        "capacity",
    ]

    def __init__(self, *args: Any, n_binding_sites: int = 2, **kwargs: Any) -> None:
        """Initialize BiLangmuirLDF."""
        self.n_binding_sites = n_binding_sites

        super().__init__(*args, **kwargs)


class FreundlichLDF(BindingBaseClass):
    """
    Freundlich isotherm model.

    Attributes
    ----------
    driving_force_coefficient : list of unsigned floats.
        Driving force coefficient for each component. Length depends on `n_comp`.
    freundlich_coefficient : list of unsigned floats.
        Freundlich coefficient for each component. Length depends on `n_comp`.
    exponent : list of unsigned floats.
        Freundlich exponent for each component. Length depends on `n_comp`.
    """

    driving_force_coefficient = SizedUnsignedList(size="n_comp")
    freundlich_coefficient = SizedUnsignedList(size="n_comp")
    exponent = SizedUnsignedList(size="n_comp")

    _parameters = [
        "driving_force_coefficient",
        "freundlich_coefficient",
        "exponent",
    ]


class StericMassAction(BindingBaseClass):
    r"""
    Steric Mass Action Law binding model.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants. Length depends on `n_comp`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants. Length depends on `n_comp`.
    characteristic_charge : list of unsigned floats.
        Characteristic charges of the protein; The number of sites $\nu$ that the
        protein interacts with on the resin surface.
        Length depends on `n_comp`.
    steric_factor : list of unsigned floats.
        Steric factors of the protein: The number of sites $\sigma$ on the surface
        that are shielded by the protein and prevented from exchange with salt
        counterions in solution.
        Length depends on `n_comp`.
    capacity : unsigned float.
        Stationary phase capacity (monovalent salt counterions); The total
        number of binding sites available on the resin surface.
    reference_liquid_phase_conc : unsigned float.
        Reference liquid phase concentration.
        The default is 1.0
    reference_solid_phase_conc : unsigned float.
        Reference liquid phase concentration.
        The default is 1.0
    """

    adsorption_rate = SizedUnsignedList(size="n_comp")
    desorption_rate = SizedUnsignedList(size="n_comp")
    characteristic_charge = SizedUnsignedList(size="n_comp")
    steric_factor = SizedUnsignedList(size="n_comp")
    capacity = UnsignedFloat()
    reference_liquid_phase_conc = UnsignedFloat(default=1.0)
    reference_solid_phase_conc = UnsignedFloat(default=1.0)

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
        "characteristic_charge",
        "steric_factor",
        "capacity",
        "reference_liquid_phase_conc",
        "reference_solid_phase_conc",
    ]

    @property
    def adsorption_rate_untransformed(self) -> list[UnsignedFloat]:
        """list[float]: Untransformed adsorption rate."""
        if self.adsorption_rate is None:
            return None

        nu = np.array(self.characteristic_charge)
        return self.adsorption_rate * self.reference_solid_phase_conc ** (-nu)

    @adsorption_rate_untransformed.setter
    def adsorption_rate_untransformed(
        self,
        adsorption_rate_untransformed: list[float],
    ) -> None:
        if self.characteristic_charge is None:
            raise ValueError(
                "Please set nu before setting an untransformed rate constant."
            )

        nu = np.array(self.characteristic_charge)
        self.adsorption_rate = (
            adsorption_rate_untransformed / self.reference_solid_phase_conc ** (-nu)
        ).tolist()

    @property
    def desorption_rate_untransformed(self) -> list[UnsignedFloat]:
        """list[float]: Untransformed desorption rate."""
        if self.desorption_rate is None:
            return None

        nu = np.array(self.characteristic_charge)
        return self.desorption_rate * self.reference_liquid_phase_conc ** (-nu)

    @desorption_rate_untransformed.setter
    def desorption_rate_untransformed(
        self,
        desorption_rate_untransformed: list[float],
    ) -> None:
        if self.characteristic_charge is None:
            raise ValueError(
                "Please set nu before setting a transformed rate constant."
            )

        nu = np.array(self.characteristic_charge)
        self.desorption_rate = (
            desorption_rate_untransformed / self.reference_liquid_phase_conc ** (-nu)
        ).tolist()


class AntiLangmuir(BindingBaseClass):
    """
    Multi Component Anti-Langmuir adsorption isotherm.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants. Length depends on `n_comp`.
    desorption_rate : list of unsigned floats
        Desorption rate constants. Length depends on `n_comp`.
    capacity : list of unsigned floats.
        Maximum adsorption capacities. Length depends on `n_comp`.
    antilangmuir : list of {-1, 1}.
        Anti-Langmuir coefficients. Length depends on `n_comp`.
    """

    adsorption_rate = SizedUnsignedList(size="n_comp")
    desorption_rate = SizedUnsignedList(size="n_comp")
    capacity = SizedUnsignedList(size="n_comp")
    antilangmuir = SizedFloatList(size="n_comp")

    _parameters = ["adsorption_rate", "desorption_rate", "capacity", "antilangmuir"]


class Spreading(BindingBaseClass):
    """
    Multi Component Spreading adsorption isotherm.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants in state-major ordering.
        Length depends on `n_bound_states`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants in state-major ordering.
        Length depends on `n_bound_states`.
    capacity : list of unsigned floats.
        Maximum adsorption capacities in state-major ordering.
        Length depends on `n_bound_states`.
    exchange_from_1_2 : list of unsigned floats.
        Exchange rates from the first to the second bound state.
        Length depends on `n_comp`.
    exchange_from_2_1 : list of unsigned floats.
        Exchange rates from the second to the first bound state.
        Length depends on `n_comp`.
    """

    n_binding_sites = RangedInteger(lb=2, ub=2, default=2)

    adsorption_rate = SizedUnsignedList(size="n_bound_states")
    desorption_rate = SizedUnsignedList(size="n_bound_states")
    capacity = SizedUnsignedList(size="n_bound_states")
    exchange_from_1_2 = SizedUnsignedList(size="n_comp")
    exchange_from_2_1 = SizedUnsignedList(size="n_comp")

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
        "capacity",
        "exchange_from_1_2",
        "exchange_from_2_1",
    ]


class MobilePhaseModulator(BindingBaseClass):
    """
    Mobile Phase Modulator adsorption isotherm.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants. Length depends on `n_comp`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants. Length depends on `n_comp`.
    capacity : list of unsigned floats.
        Maximum adsorption capacities. Length depends on `n_comp`.
    ion_exchange_characteristic : list of unsigned floats.
        Parameters describing the ion-exchange characteristics (IEX).
        Length depends on `n_comp`.
    hydrophobicity : list of floats.
        Parameters describing the hydrophobicity (HIC).
        Length depends on `n_comp`.
    linear_threshold : UnsignedFloat
        Concentration of c0 at which to switch to linear model approximation.
    """

    adsorption_rate = SizedUnsignedList(size="n_comp")
    desorption_rate = SizedUnsignedList(size="n_comp")
    capacity = SizedUnsignedList(size="n_comp")
    ion_exchange_characteristic = SizedUnsignedList(size="n_comp")
    beta = ion_exchange_characteristic
    hydrophobicity = SizedFloatList(size="n_comp")
    gamma = hydrophobicity
    linear_threshold = UnsignedFloat(default=1e-8)

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
        "capacity",
        "ion_exchange_characteristic",
        "hydrophobicity",
        "linear_threshold",
    ]


class ExtendedMobilePhaseModulator(BindingBaseClass):
    """
    Mobile Phase Modulator adsorption isotherm.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants. Length depends on `n_comp`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants. Length depends on `n_comp`.
    capacity : list of unsigned floats.
        Maximum adsorption capacities. Length depends on `n_comp`.
    ion_exchange_characteristic : list of unsigned floats.
        Parameters describing the ion-exchange characteristics (IEX).
        Length depends on `n_comp`.
    hydrophobicity : list of floats.
        Parameters describing the hydrophobicity (HIC).
        Length depends on `n_comp`.
    component_mode : list of unsigned integers.
        Mode of each component;
        0 denotes the modifier component,
        1 is linear binding,
        2 is modified Langmuir binding.
        Length depends on `n_comp`.
    """

    adsorption_rate = SizedUnsignedList(size="n_comp")
    desorption_rate = SizedUnsignedList(size="n_comp")
    capacity = SizedUnsignedList(size="n_comp")
    ion_exchange_characteristic = SizedUnsignedList(size="n_comp")
    beta = ion_exchange_characteristic
    hydrophobicity = SizedFloatList(size="n_comp")
    gamma = hydrophobicity
    component_mode = SizedUnsignedIntegerList(size="n_comp", ub=2)

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
        "capacity",
        "ion_exchange_characteristic",
        "hydrophobicity",
        "component_mode",
    ]


class SelfAssociation(BindingBaseClass):
    r"""
    Self Association adsorption isotherm.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants. Length depends on `n_comp`.
    adsorption_rate_dimerization : list of unsigned floats.
        Adsorption rate constants of dimerization. Length depends on `n_comp`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants. Length depends on `n_comp`.
    characteristic_charge : list of unsigned floats.
        The characteristic charge $\nu$ of the protein. Length depends on `n_comp`.
    steric_factor : list of unsigned floats.
        Steric factor of of the protein. Length depends on `n_comp`.
    capacity : unsigned float.
        Stationary phase capacity (monovalent salt counterions); The total number of
        binding sites available on the resin surface.
    reference_liquid_phase_conc : unsigned float.
        Reference liquid phase concentration (optional, default value = 1.0).
        The default = 1.0
    reference_solid_phase_conc : unsigned float.
        Reference liquid phase concentration (optional)
        The default = 1.0
    """

    adsorption_rate = SizedUnsignedList(size="n_comp")
    adsorption_rate_dimerization = SizedUnsignedList(size="n_comp")
    desorption_rate = SizedUnsignedList(size="n_comp")
    characteristic_charge = SizedUnsignedList(size="n_comp")
    steric_factor = SizedUnsignedList(size="n_comp")
    capacity = UnsignedFloat()
    reference_liquid_phase_conc = UnsignedFloat(default=1.0)
    reference_solid_phase_conc = UnsignedFloat(default=1.0)

    _parameters = [
        "adsorption_rate",
        "adsorption_rate_dimerization",
        "desorption_rate",
        "characteristic_charge",
        "steric_factor",
        "capacity",
        "reference_liquid_phase_conc",
        "reference_solid_phase_conc",
    ]


class BiStericMassAction(BindingBaseClass):
    """
    Bi Steric Mass Action adsorption isotherm.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants in state-major ordering.
        Length depends on `n_bound_states`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants in state-major ordering.
        Length depends on `n_bound_states`.
    characteristic_charge : list of unsigned floats.
        Characteristic charges v(i,j) of the it-h protein with respect to the
        j-th binding site type in state-major ordering.
        Length depends on `n_bound_states`.
    steric_factor : list of unsigned floats.
        Steric factor o (i,j) of the it-h protein with respect to the j-th
        binding site type in state-major ordering.
        Length depends on `n_bound_states`.
    capacity : unsigned float.
        Stationary phase capacity (monovalent salt counterions); The total
        number of binding sites available on the resin surface.
        Length depends on `n_binding_sites`.
    reference_liquid_phase_conc : list of unsigned floats.
        Reference liquid phase concentration for each binding site type or one
        value for all types.
        The default is 1.0
    reference_solid_phase_conc : list of unsigned floats.
        Reference solid phase concentration for each binding site type or one
        value for all types.
        The default is 1.0
    """

    n_binding_sites = UnsignedInteger(default=2)

    adsorption_rate = SizedUnsignedList(size="n_bound_states")
    adsorption_rate_dimerization = SizedUnsignedList(size="n_bound_states")
    desorption_rate = SizedUnsignedList(size="n_bound_states")
    characteristic_charge = SizedUnsignedList(size="n_bound_states")
    steric_factor = SizedUnsignedList(size="n_bound_states")
    capacity = SizedUnsignedList(size="n_binding_sites")
    reference_liquid_phase_conc = SizedUnsignedList(size="n_binding_sites", default=1)
    reference_solid_phase_conc = SizedUnsignedList(size="n_binding_sites", default=1)

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
        "characteristic_charge",
        "steric_factor",
        "capacity",
        "reference_liquid_phase_conc",
        "reference_solid_phase_conc",
    ]

    def __init__(self, *args: Any, n_states: int = 2, **kwargs: Any) -> None:
        """Initialize BiStericMassAction class."""
        self.n_states = n_states
        super().__init__(*args, **kwargs)


class MultistateStericMassAction(BindingBaseClass):
    r"""
    Multistate Steric Mass Action adsorption isotherm.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants of the components to different bound states in
        component-major ordering.
        Length depends on `n_bound_states`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants of the components to different bound states in
        component-major ordering.
        Length depends on `n_bound_states`.
    characteristic_charge : list of unsigned floats.
        Characteristic charges of the components to different bound states in
        component-major ordering.
        Length depends on `n_bound_states`.
    steric_factor : list of unsigned floats.
        Steric factor of the components to different bound states in component-major
        ordering.
        Length depends on `n_bound_states`.
    conversion_rate : list of unsigned floats.
        Conversion rates between different bound states in
        component-major ordering.
        Length: $sum_{i=1}^{n_{comp}} n_{bound, i}$
    capacity : unsigned float.
        Stationary phase capacity (monovalent salt counterions); The total number of
        binding sites available on the resin surface.
    reference_liquid_phase_conc : unsigned float.
        Reference liquid phase concentration.
        The default = 1.0
    reference_solid_phase_conc : unsigned float, optional
        Reference solid phase concentration.
        The default = 1.0
    """

    bound_states = SizedUnsignedIntegerList(
        size=("n_binding_sites", "n_comp"), default=1
    )

    adsorption_rate = SizedUnsignedList(size="n_bound_states")
    desorption_rate = SizedUnsignedList(size="n_bound_states")
    characteristic_charge = SizedUnsignedList(size="n_bound_states")
    steric_factor = SizedUnsignedList(size="n_bound_states")
    conversion_rate = SizedUnsignedList(size="_conversion_entries")
    capacity = UnsignedFloat()
    reference_liquid_phase_conc = UnsignedFloat(default=1)
    reference_solid_phase_conc = UnsignedFloat(default=1)

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
        "characteristic_charge",
        "steric_factor",
        "conversion_rate",
        "capacity",
        "reference_liquid_phase_conc",
        "reference_solid_phase_conc",
    ]

    @property
    def _conversion_entries(self) -> int:
        n = 0
        for state in self.bound_states:
            n += state**2

        return n


class SimplifiedMultistateStericMassAction(BindingBaseClass):
    """
    Simplified multistate Steric Mass Action adsorption isotherm.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants of the components to different bound states
        in component-major ordering.
        Length depends on `n_bound_states`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants of the components to different bound states
        in component-major ordering.
        Length depends on `n_bound_states`.
    characteristic_charge_first : list of unsigned floats.
        Characteristic charges of the components in the first (weakest) bound state.
        Length depends on `n_comp`.
    characteristic_charge_last : list of unsigned floats.
        Characteristic charges of the components in the last (strongest) bound state.
        Length depends on `n_comp`.
    quadratic_modifiers_charge : list of unsigned floats.
        Quadratic modifiers of the characteristic charges of the different components
        depending on the index of the bound state.
        Length depends on `n_comp`.
    steric_factor_first : list of unsigned floats.
        Steric factor of the components in the first (weakest) bound state.
        Length depends on `n_comp`.
    steric_factor_last : list of unsigned floats.
        Steric factor of the components in the last (strongest) bound state.
        Length depends on `n_comp`.
    quadratic_modifiers_steric : list of unsigned floats.
        Quadratic modifiers of the sterif factors of the different components depending
        on the index of the bound state.
        Length depends on `n_comp`.
    capacity : unsigned float.
        Stationary phase capacity (monovalent salt counterions): The total number of
        binding sites available on the resin surface.
    exchange_from_weak_stronger : list of unsigned floats.
        Exchangde rated from a weakly bound state to the next stronger bound state.
    linear_exchange_ws : list of unsigned floats.
        Linear exchange rate coefficients from a weakly bound state to the next stronger
        bound state.
        Length depends on `n_comp`.
    quadratic_exchange_ws : list of unsigned floats.
        Quadratic exchange rate coefficients from a weakly bound state to the next
        stronger bound state.
        Length depends on `n_comp`.
    exchange_from_stronger_weak : list of unsigned floats.
        Exchange rate coefficients from a strongly bound state to the next weaker bound
        state.
        Length depends on `n_comp`.
    linear_exchange_sw : list of unsigned floats.
        Linear exchange rate coefficients from a strongly bound state to the next weaker
        bound state.
        Length depends on `n_comp`.
    quadratic_exchange_sw : list of unsigned floats.
        Quadratic exchange rate coefficients from a strongly bound state to the next
        weaker bound state.
        Length depends on `n_comp`.
    reference_liquid_phase_conc : list of unsigned floats.
        Reference liquid phase concentration (optional, default value = 1.0).
    reference_solid_phase_conc : list of unsigned floats.
        Reference solid phase concentration (optional, default value = 1.0).
    """

    bound_states = SizedUnsignedIntegerList(
        size=("n_binding_sites", "n_comp"), default=1
    )

    adsorption_rate = SizedUnsignedList(size="n_bound_states")
    desorption_rate = SizedUnsignedList(size="n_bound_states")
    characteristic_charge_first = SizedUnsignedList(size="n_comp")
    characteristic_charge_last = SizedUnsignedList(size="n_comp")
    quadratic_modifiers_charge = SizedUnsignedList(size="n_comp")
    steric_factor_first = SizedUnsignedList(size="n_comp")
    steric_factor_last = SizedUnsignedList(size="n_comp")
    quadratic_modifiers_steric = SizedUnsignedList(size="n_comp")
    capacity = UnsignedFloat()
    exchange_from_weak_stronger = SizedUnsignedList(size="n_comp")
    linear_exchange_ws = SizedUnsignedList(size="n_comp")
    quadratic_exchange_ws = SizedUnsignedList(size="n_comp")
    exchange_from_stronger_weak = SizedUnsignedList(size="n_comp")
    linear_exchange_sw = SizedUnsignedList(size="n_comp")
    quadratic_exchange_sw = SizedUnsignedList(size="n_comp")
    reference_liquid_phase_conc = UnsignedFloat(default=1)
    reference_solid_phase_conc = UnsignedFloat(default=1)

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
        "characteristic_charge_first",
        "characteristic_charge_last",
        "quadratic_modifiers_charge",
        "steric_factor_first",
        "steric_factor_last",
        "quadratic_modifiers_steric",
        "capacity",
        "exchange_from_weak_stronger",
        "linear_exchange_ws",
        "quadratic_exchange_ws",
        "exchange_from_stronger_weak",
        "linear_exchange_sw",
        "quadratic_exchange_sw",
        "reference_liquid_phase_conc",
        "reference_solid_phase_conc",
    ]


class Saska(BindingBaseClass):
    """
    Quadratic Isotherm.

    Attributes
    ----------
    henry_const : list of unsigned floats.
        The Henry coefficient. Length depends on `n_comp`.
    quadratic_factor : list of unsigned floats.
        Quadratic factors. Length depends on `n_comp`.
    """

    henry_const = SizedUnsignedList(size="n_comp")
    quadratic_factor = SizedUnsignedList(size=("n_comp", "n_comp"))

    _parameters = [
        "henry_const",
        "quadratic_factor",
    ]


class GeneralizedIonExchange(BindingBaseClass):
    r"""
    Generalized Ion Exchange isotherm model.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants. Length depends on `n_comp`.
    adsorption_rate_linear : list of unsigned floats.
        Linear dependence coefficient of adsorption rate on modifier component
        Length depends on `n_comp`.
    adsorption_rate_quadratic : list of unsigned floats.
        Quadratic dependence coefficient of adsorption rate on modifier component.
        Length depends on `n_comp`.
    adsorption_rate_cubic : list of unsigned floats.
        Cubic dependence coefficient of adsorption rate on modifier component.
        Length depends on `n_comp`.
    adsorption_rate_salt : list of unsigned floats.
        Salt coefficient of adsorption rate;
        difference of water-protein and salt-protein interactions.
        Length depends on `n_comp`.
    adsorption_rate_protein : list of unsigned floats.
        Protein coefficient of adsorption rate;
        difference of water-protein and protein-protein interactions.
        Length depends on `n_comp`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants. Length depends on `n_comp`.
    desorption_rate_linear : list of unsigned floats.
        Linear dependence coefficient of desorption rate on modifier component.
        Length depends on `n_comp`.
    desorption_rate_quadratic : list of unsigned floats.
        Quadratic dependence coefficient of desorption rate on modifier component.
        Length depends on `n_comp`.
    desorption_rate_cubic : list of unsigned floats.
        Cubic dependence coefficient of desorption rate on modifier component.
        Length depends on `n_comp`.
    desorption_rate_salt : list of unsigned floats.
        Salt coefficient of desorption rate;
        difference of water-protein and salt-protein interactions.
        Length depends on `n_comp`.
    desorption_rate_protein : list of unsigned floats.
        Protein coefficient of desorption rate;
        difference of water-protein and protein-protein interactions
        Length depends on `n_comp`.
    characteristic_charge_breaks : list of unsigned floats, optional
        Breaks of the characteristic charge pieces in component-major ordering.
        Optional, only required if a piecewise cubic polynomial is used for $\nu$.
        Length must be a multiple of `n_comp`.
    characteristic_charge : list of unsigned floats.
        Base value for characteristic charges of the protein; The number of sites $\nu$
        that the protein interacts with on the resin surface.
        Length depends on `n_comp` * `n_pieces`.
    characteristic_charge_linear : list of unsigned floats.
        Linear dependence coefficient of characteristic charge on modifier component.
        Length depends on `n_comp` * `n_pieces`.
    characteristic_charge_quadratic : list of unsigned floats.
        Quadratic dependence coefficient of characteristic charge on modifier component.
        Length depends on `n_comp` * `n_pieces`.
    characteristic_charge_cubic : list of unsigned floats.
        Cubic dependence coefficient of characteristic charge on modifier component.
        Length depends on `n_comp` * `n_pieces`.
    steric_factor : list of unsigned floats.
        Steric factors of the protein: The number of sites $\sigma$ on the surface that
        are shielded by the protein and prevented from exchange with salt counterions in
        solution.
        Length depends on `n_comp`.
    capacity : unsigned float.
        Stationary phase capacity (monovalent salt counterions); The total number of
        binding sites available on the resin surface.
    reference_liquid_phase_conc : unsigned float.
        Reference liquid phase concentration (optional, default value = 1.0).
    reference_solid_phase_conc : unsigned float.
        Reference liquid phase concentration (optional, default value = 1.0).
    """

    non_binding_component_indices = [1]

    adsorption_rate = SizedFloatList(size="n_comp")
    adsorption_rate_linear = SizedFloatList(size="n_comp")
    adsorption_rate_quadratic = SizedFloatList(size="n_comp", default=0)
    adsorption_rate_cubic = SizedFloatList(size="n_comp", default=0)
    adsorption_rate_salt = SizedFloatList(size="n_comp", default=0)
    adsorption_rate_protein = SizedFloatList(size="n_comp", default=0)
    desorption_rate = SizedFloatList(size="n_comp")
    desorption_rate_linear = SizedFloatList(size="n_comp", default=0)
    desorption_rate_quadratic = SizedFloatList(size="n_comp", default=0)
    desorption_rate_cubic = SizedFloatList(size="n_comp", default=0)
    desorption_rate_salt = SizedFloatList(size="n_comp", default=0)
    desorption_rate_protein = SizedFloatList(size="n_comp", default=0)
    characteristic_charge_breaks = DependentlyModulatedUnsignedList(
        size="n_comp", is_optional=True
    )
    characteristic_charge = SizedFloatList(
        size=("n_pieces", "n_comp"),
    )
    characteristic_charge_linear = SizedFloatList(
        size=("n_pieces", "n_comp"), default=0
    )
    characteristic_charge_quadratic = SizedFloatList(
        size=("n_pieces", "n_comp"), default=0
    )
    characteristic_charge_cubic = SizedFloatList(size=("n_pieces", "n_comp"), default=0)
    steric_factor = SizedUnsignedList(size="n_comp")
    capacity = UnsignedFloat()
    reference_liquid_phase_conc = UnsignedFloat(default=1)
    reference_solid_phase_conc = UnsignedFloat(default=1)

    _parameters = [
        "adsorption_rate",
        "adsorption_rate_linear",
        "adsorption_rate_quadratic",
        "adsorption_rate_cubic",
        "adsorption_rate_salt",
        "adsorption_rate_protein",
        "desorption_rate",
        "desorption_rate_linear",
        "desorption_rate_quadratic",
        "desorption_rate_cubic",
        "desorption_rate_salt",
        "desorption_rate_protein",
        "characteristic_charge_breaks",
        "characteristic_charge",
        "characteristic_charge_linear",
        "characteristic_charge_quadratic",
        "characteristic_charge_cubic",
        "steric_factor",
        "capacity",
        "reference_liquid_phase_conc",
        "reference_solid_phase_conc",
    ]

    @property
    def n_pieces(self) -> int:
        """int: Number of pieces for cubic polynomial description of nu."""
        if self.characteristic_charge_breaks is None:
            return 1

        n_pieces_all = len(self.characteristic_charge_breaks) - self.n_comp
        n_pieces_comp = int(n_pieces_all / self.n_comp)

        return n_pieces_comp


class HICConstantWaterActivity(BindingBaseClass):
    """
    HIC based on Constant Water Activity adsorption isotherm.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants. Size depends on `n_comp`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants. Size depends on `n_comp`.
    capacity : list of unsigned floats.
        Maximum adsorption capacities. Size depends on `n_comp`.
    hic_characteristic : list of unsigned floats.
        Parameters describing the number of ligands per ligand-protein interaction.
        Size depends on `n_comp`.
    beta_0 : unsigned float.
        Parameter describing the number of highly ordered water molecules that stabilize
        the hydrophobic surfaces at infinitely diluted salt concentration.
    beta_1 : unsigned float.
        Parameter describing the change in the number of highly ordered water molecules that
        stabilize the hydrophobic surfaces with respect to changes in the salt concentration.
    """

    adsorption_rate = SizedFloatList(size="n_comp")
    desorption_rate = SizedFloatList(size="n_comp")
    capacity = SizedFloatList(size="n_comp")
    hic_characteristic = SizedFloatList(size="n_comp")

    beta_0 = UnsignedFloat()
    beta_1 = UnsignedFloat()

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
        "hic_characteristic",
        "capacity",
        "beta_0",
        "beta_1",
    ]


class HICWaterOnHydrophobicSurfaces(BindingBaseClass):
    """
    HIC isotherm by Wang et al. based on their 2016 paper.

    Attributes
    ----------
    adsorption_rate : list of unsigned floats.
        Adsorption rate constants. Size depends on `n_comp`.
    desorption_rate : list of unsigned floats.
        Desorption rate constants. Size depends on `n_comp`.
    capacity : list of unsigned floats.
        Maximum adsorption capacities. Size depends on `n_comp`.
    hic_characteristic : list of unsigned floats.
        Parameters describing the number of ligands per ligand-protein interaction.
        Size depends on `n_comp`.
    beta_0 : unsigned float.
        Parameter describing the number of highly ordered water molecules that stabilize
        the hydrophobic surfaces at infinitely diluted salt concentration.
    beta_1 : unsigned float.
        Parameter describing the change in the number of highly ordered water molecules that
        stabilize the hydrophobic surfaces with respect to changes in the salt concentration.
    """

    adsorption_rate = SizedFloatList(size="n_comp")
    desorption_rate = SizedFloatList(size="n_comp")
    capacity = SizedFloatList(size="n_comp")
    hic_characteristic = SizedFloatList(size="n_comp")

    beta_0 = UnsignedFloat()
    beta_1 = UnsignedFloat()

    _parameters = [
        "adsorption_rate",
        "desorption_rate",
        "hic_characteristic",
        "capacity",
        "beta_0",
        "beta_1",
    ]


class MultiComponentColloidal(BindingBaseClass):
    """
    Colloidal isotherm from Xu and Lenhoff 2009.

    Attributes
    ----------
    phase_ratio : unsigned float.
        Phase ratio.
    kappa_exponential : unsigned float.
        Screening term exponential factor.
    kappa_factor : unsigned float.
        Screening term factor.
    kappa_constant : unsigned float.
        Screening term constant.
    coordination_number : unsigned integer.
        Coordination number.
    logkeq_ph_exponent : list of unsigned floats.
        Equilibrium constant factor exponent term for pH. Size depends on `n_comp`.
    logkeq_power_exponent : list of unsigned floats.
        Equilibrium constant power exponent term for salt. Size depends on `n_comp`.
    logkeq_power_factor : list of unsigned floats.
        Equilibrium constant power factor term for salt. Size depends on `n_comp`.
    logkeq_exponent_factor : list of unsigned floats.
        Equilibrium constant exponent factor term for salt. Size depends on `n_comp`.
    logkeq_exponent_multiplier : list of unsigned floats.
        Equilibrium constant exponent multiplier term for salt. Size depends on `n_comp`.
    bpp_ph_exponent : list of unsigned floats..
        BPP constant exponent factor term for pH. Size depends on `n_comp`.
    bpp_power_exponent : list of unsigned floats.
        Bpp constant power exponent term for salt. Size depends on `n_comp`.
    bpp_power_factor : list of unsigned floats.
        Bpp constant power factor term for salt. Size depends on `n_comp`.
    bpp_exponent_factor  : list of unsigned floats.
        Bpp constant exponent factor term for salt. Size depends on `n_comp`.
    bpp_exponent_multiplier : list of unsigned floats.
        Bpp constant exponent multiplier term for salt. Size depends on `n_comp`.
    protein_radius : list of unsigned floats.
        Protein radius. Size depends on `n_comp`.
    kinetic_rate_constant : list of unsigned floats.
        Adsorption kinetics. Size depends on `n_comp`.
    linear_threshold : unsigned float.
        Linear threshold.
    use_ph : Boolean.
        Include pH or not.
    """

    bound_states = SizedUnsignedIntegerList(
        size=("n_binding_sites", "n_comp"), default=1
    )

    phase_ratio = UnsignedFloat()
    kappa_exponential = UnsignedFloat()
    kappa_factor = UnsignedFloat()
    kappa_constant = UnsignedFloat()
    coordination_number = UnsignedInteger()
    logkeq_ph_exponent = SizedFloatList(size="n_comp")
    logkeq_power_exponent = SizedFloatList(size="n_comp")
    logkeq_power_factor = SizedFloatList(size="n_comp")
    logkeq_exponent_factor = SizedFloatList(size="n_comp")
    logkeq_exponent_multiplier = SizedFloatList(size="n_comp")
    bpp_ph_exponent = SizedFloatList(size="n_comp")
    bpp_power_exponent = SizedFloatList(size="n_comp")
    bpp_power_factor = SizedFloatList(size="n_comp")
    bpp_exponent_factor = SizedFloatList(size="n_comp")
    bpp_exponent_multiplier = SizedFloatList(size="n_comp")
    protein_radius = SizedFloatList(size="n_comp")
    kinetic_rate_constant = SizedFloatList(size="n_comp")
    linear_threshold = UnsignedFloat(default=1e-8)
    use_ph = Bool(default=False)

    _parameters = [
        "phase_ratio",
        "kappa_exponential",
        "kappa_factor",
        "kappa_constant",
        "coordination_number",
        "logkeq_ph_exponent",
        "logkeq_power_exponent",
        "logkeq_power_factor",
        "logkeq_exponent_factor",
        "logkeq_exponent_multiplier",
        "bpp_ph_exponent",
        "bpp_power_exponent",
        "bpp_power_factor",
        "bpp_exponent_factor",
        "bpp_exponent_multiplier",
        "protein_radius",
        "kinetic_rate_constant",
        "linear_threshold",
        "use_ph",
    ]



import unittest

from CADETProcess.transform import (
    AutoTransformer,
    NormLinearTransformer,
    NormLogTransformer,
    NullTransformer,
)


class Test_Transformer(unittest.TestCase):
    def test_input_range(self):
        transform = NormLinearTransformer(0, 100)

        with self.assertRaises(ValueError):
            in_ = -10
            out = transform.transform(in_)

        with self.assertRaises(ValueError):
            in_ = 1000
            out = transform.transform(in_)

    def test_output_range(self):
        transform = NormLinearTransformer(0, 100)

        with self.assertRaises(ValueError):
            in_ = -1
            out = transform.untransform(in_)

        with self.assertRaises(ValueError):
            in_ = 2
            out = transform.untransform(in_)

    def test_no_transform(self):
        transform = NullTransformer(0, 100)
        self.assertAlmostEqual(transform.lb, 0)
        self.assertAlmostEqual(transform.ub, 100)

        in_ = 0
        out_expected = 0
        out = transform.transform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 0
        out_expected = 0
        out = transform.untransform(in_)
        self.assertAlmostEqual(out_expected, out)

    def test_linear(self):
        transform = NormLinearTransformer(0, 100)
        self.assertAlmostEqual(transform.lb, 0)
        self.assertAlmostEqual(transform.ub, 1)

        in_ = 0
        out_expected = 0
        out = transform.transform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 0
        out_expected = 0
        out = transform.untransform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 10
        out_expected = 0.1
        out = transform.transform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 0.1
        out_expected = 10
        out = transform.untransform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 100
        out_expected = 1
        out = transform.transform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 1
        out_expected = 100
        out = transform.untransform(in_)
        self.assertAlmostEqual(out_expected, out)

    def test_log(self):
        """Missing: Special case for lb_input <= 0"""
        transform = NormLogTransformer(1, 1000)
        self.assertAlmostEqual(transform.lb, 0)
        self.assertAlmostEqual(transform.ub, 1)

        in_ = 1
        out_expected = 0
        out = transform.transform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 0
        out_expected = 1
        out = transform.untransform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 10
        out_expected = 1 / 3
        out = transform.transform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 1 / 3
        out_expected = 10
        out = transform.untransform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 100
        out_expected = 2 / 3
        out = transform.transform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 2 / 3
        out_expected = 100
        out = transform.untransform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 1000
        out_expected = 1
        out = transform.transform(in_)
        self.assertAlmostEqual(out_expected, out)

        in_ = 1
        out_expected = 1000
        out = transform.untransform(in_)
        self.assertAlmostEqual(out_expected, out)

    def test_auto(self):
        threshold = 1000

        transform = AutoTransformer(1, 100, threshold=threshold)
        self.assertTrue(transform.use_linear)
        self.assertFalse(transform.use_log)

        # Expect Log behaviour
        transform = AutoTransformer(1, 1001, threshold=threshold)
        self.assertFalse(transform.use_linear)
        self.assertTrue(transform.use_log)


if __name__ == "__main__":
    unittest.main()



from functools import partial

import numpy as np
import pytest
from CADETProcess.optimization import (
    COBYLA,
    SLSQP,
    U_NSGA3,
    NelderMead,
    OptimizerBase,
    TrustConstr,
)

skip_ax = False
try:
    from CADETProcess.optimization import (
        GPEI,
        NEHVI,
        qNParEGO,
    )
except ImportError:
    skip_ax = True

from tests.optimization_problem_fixtures import (
    LinearConstraintsMooTestProblem,
    LinearConstraintsSooTestProblem,
    LinearConstraintsSooTestProblem2,
    LinearEqualityConstraintsSooTestProblem,
    LinearNonlinearConstraintsMooTestProblem,
    NonlinearConstraintsMooTestProblem,
    NonlinearConstraintsSooTestProblem,
    NonlinearLinearConstraintsSooTestProblem,
    Rosenbrock,
    TestProblem,
)  # noqa: E402

# %% Optimizer Setup

SOO_TEST_KWARGS = {
    "atol": 0.05,   # allows absolute 0.05 deviation (low values) of solution or
    "rtol": 0.1,    # 0.1 alows 10% deviation of true solution
}

MOO_TEST_KWARGS = {
    "atol": 0.01,
    "rtol": 0.1,
    "mismatch_tol": 0.3333333333,  # 75 % of all solutions must lie on the pareto front
}

F_TOL = 0.001
X_TOL = 0.001

CV_BOUNDS_TOL = 1e-9
CV_LINCON_TOL = 1e-6
CV_LINEQCON_TOL = 1e-9
CV_NONLINCON_TOL = 0.0001

EXCLUDE_COMBINATIONS = [
    (
        U_NSGA3,
        LinearEqualityConstraintsSooTestProblem,
        "See also: https://jugit.fz-juelich.de/IBG-1/ModSim/hopsy/-/issues/152",
    )
]

# this helps to test optimizers for hard problems
NON_DEFAULT_PARAMETERS = [
    (U_NSGA3, NonlinearConstraintsMooTestProblem, {"pop_size": 300, "n_max_gen": 40}),
    (U_NSGA3, Rosenbrock, {"pop_size": 300, "n_max_gen": 20}),
]


def skip_if_combination_excluded(optimizer, problem):
    for o, p, r in EXCLUDE_COMBINATIONS:
        if isinstance(optimizer, o) and isinstance(problem, p):
            pytest.skip(reason=r)


def set_non_default_parameters(optimizer, problem):
    for o, p, params in NON_DEFAULT_PARAMETERS:
        if isinstance(optimizer, o) and isinstance(problem, p):
            # override default parameters of the optimizer
            for pk, pv in params.items():
                setattr(optimizer, pk, pv)


class TrustConstr(TrustConstr):
    x_tol = X_TOL
    cv_nonlincon_tol = CV_NONLINCON_TOL


class COBYLA(COBYLA):
    x_tol = X_TOL
    cv_nonlincon_tol = CV_NONLINCON_TOL


class NelderMead(NelderMead):
    x_tol = X_TOL
    f_tol = F_TOL


class SLSQP(SLSQP):
    x_tol = X_TOL
    cv_lincon_tol = CV_LINCON_TOL


class U_NSGA3(U_NSGA3):
    f_tol = F_TOL
    x_tol = X_TOL
    cv_nonlincon_tol = CV_NONLINCON_TOL
    pop_size = 100
    n_max_gen = 20  # before used 100 generations --> this did not improve the fit


if not skip_ax:

    class GPEI(GPEI):
        cv_lincon_tol = CV_LINCON_TOL
        n_init_evals = 40
        early_stopping_improvement_bar = 1e-4
        early_stopping_improvement_window = 10
        n_max_evals = 50

    class NEHVI(NEHVI):
        cv_lincon_tol = CV_LINCON_TOL
        n_init_evals = 50
        early_stopping_improvement_bar = 1e-4
        early_stopping_improvement_window = 10
        n_max_evals = 60

    class qNParEGO(qNParEGO):
        cv_lincon_tol = CV_LINCON_TOL
        n_init_evals = 50
        early_stopping_improvement_bar = 1e-4
        early_stopping_improvement_window = 10
        n_max_evals = 70


# %% Test problem factory


@pytest.fixture(
    params=[
        # single objective problems
        Rosenbrock,
        LinearConstraintsSooTestProblem,
        LinearConstraintsSooTestProblem2,
        NonlinearConstraintsSooTestProblem,
        LinearEqualityConstraintsSooTestProblem,
        NonlinearLinearConstraintsSooTestProblem,
        # multi objective problems
        LinearConstraintsMooTestProblem,
        NonlinearConstraintsMooTestProblem,
        LinearNonlinearConstraintsMooTestProblem,
        # transformed problems
        partial(LinearConstraintsSooTestProblem, transform="linear"),
        partial(LinearEqualityConstraintsSooTestProblem, transform="linear"),
        partial(NonlinearLinearConstraintsSooTestProblem, transform="linear"),
    ]
)
def optimization_problem(request):
    return request.param(use_diskcache=False)


params = [
    TrustConstr,
    COBYLA,
    SLSQP,
    NelderMead,
    U_NSGA3,
]

if not skip_ax:
    params.extend(
        [
            GPEI,
            NEHVI,
            qNParEGO,
        ]
    )
    EXCLUDE_COMBINATIONS.append(
        (GPEI, Rosenbrock, "cannot solve problem with enough accuracy fast enough.")
    )

    # this helps to test optimizers for hard problems
    NON_DEFAULT_PARAMETERS.append(
        (NEHVI, LinearConstraintsMooTestProblem, {"n_init_evals": 20, "n_max_evals": 40})
    )


@pytest.fixture(params=params)
def optimizer(request):
    optimizer = request.param()
    optimizer.progress_freqency = None
    return optimizer


# %% Tests
@pytest.mark.slow
def test_convergence(optimization_problem: TestProblem, optimizer: OptimizerBase):
    # only test problems that the optimizer can handle. The rest of the tests
    # will be marked as passed
    pytest.skip()

    if optimizer.check_optimization_problem(optimization_problem):
        results = optimizer.optimize(
            optimization_problem=optimization_problem,
            save_results=False,
        )
        if optimization_problem.n_objectives == 1:
            optimization_problem.test_if_solved(results, SOO_TEST_KWARGS)
        else:
            optimization_problem.test_if_solved(results, MOO_TEST_KWARGS)


@pytest.mark.slow
def test_from_initial_values(
    optimization_problem: TestProblem, optimizer: OptimizerBase
):
    if optimizer.check_optimization_problem(optimization_problem):
        skip_if_combination_excluded(optimizer, optimization_problem)
        set_non_default_parameters(optimizer, optimization_problem)

        results = optimizer.optimize(
            optimization_problem=optimization_problem,
            x0=optimization_problem.x0,
            save_results=False,
        )
        if optimization_problem.n_objectives == 1:
            optimization_problem.test_if_solved(results, SOO_TEST_KWARGS)
        else:
            optimization_problem.test_if_solved(results, MOO_TEST_KWARGS)


class AbortingCallback:
    """A callback that raises an exception after a specified number of calls."""

    def __init__(self, n_max_evals=2, abort=True):
        """Initialize callback with maximum number of evaluations and whether to abort."""
        self.n_calls = 0
        self.n_max_evals = n_max_evals
        self.abort = abort

    def reset(self):
        """Reset the number of calls to zero."""
        self.n_calls = 0

    def __call__(self, results):
        """Check the number of calls and raises an exception if the maximum has been reached."""
        if self.abort and self.n_calls >= self.n_max_evals:
            raise RuntimeError("Max number of evaluations reached. Aborting!")
        self.n_calls += 1


@pytest.mark.slow
def test_resume_from_checkpoint(
    optimization_problem: TestProblem, optimizer: OptimizerBase
):
    pytest.skip()

    # TODO: Do we need to run this for all problems?
    if optimizer.check_optimization_problem(optimization_problem):
        callback = AbortingCallback(n_max_evals=2, abort=True)
        optimization_problem.add_callback(callback)

        # TODO: How would this work for evaluation based optimizers (vs generation based)?
        optimizer.n_max_iter = 5
        if optimizer.is_population_based:
            optimizer.pop_size = 3

        try:
            opt_results = optimizer.optimize(
                optimization_problem,
                save_results=True,
                use_checkpoint=False,
            )
        except RuntimeError as e:
            if str(e) == "Max number of evaluations reached. Aborting!":
                pass
            else:
                raise e

        results_aborted = optimizer.results
        assert results_aborted.n_gen == 3
        assert callback.n_calls == 2

        callback.reset()
        callback.abort = False

        results_full = optimizer.optimize(
            optimization_problem,
            save_results=True,
            use_checkpoint=True,
        )

        # Assert all results are present
        assert results_full.n_gen == 5
        # np.testing.assertal
        np.testing.assert_almost_equal(
            results_full.populations[1].x, results_aborted.populations[1].x
        )
        # Assert callback was only called 3 times
        assert callback.n_calls == 3


if __name__ == "__main__":
    pytest.main([__file__])



import unittest

import numpy as np
from CADETProcess import CADETProcessError
from CADETProcess.processModel import (
    MCT,
    ComponentSystem,
    Cstr,
    FlowSheet,
    Inlet,
    LumpedRateModelWithoutPores,
    Outlet,
)


def assert_almost_equal_dict(dict_actual, dict_expected, decimal=7, verbose=True):
    """Helper function to assert nested dicts are (almost) equal.

    Because of floating point calculations, it is necessary to use
    `np.assert_almost_equal` to check the flow rates. However, this does not
    work well with nested dicts which is why this helper function was written.

    Parameters
    ----------
    dict_actual : dict
        The object to check.
    dict_expected : dict
        The expected object.
    decimal : int, optional
        Desired precision, default is 7.
    err_msg : str, optional
        The error message to be printed in case of failure.
    verbose : bool, optional
        If True, the conflicting values are appended to the error message.

    """
    for key in dict_actual:
        if isinstance(dict_actual[key], dict):
            assert_almost_equal_dict(dict_actual[key], dict_expected[key])

        else:
            np.testing.assert_almost_equal(
                dict_actual[key],
                dict_expected[key],
                decimal=decimal,
                err_msg=f"Dicts are not equal in key {key}.",
                verbose=verbose,
            )


def setup_single_cstr_flow_sheet(component_system=None):
    """
    Set up a simple `FlowSheet` with a single Continuous Stirred Tank Reactor (CSTR).

    Parameters
    ----------
    component_system : ComponentSystem, optional
        The component system for the CSTR.
        Defaults to a system with two components if None.

    Returns
    -------
    FlowSheet
        A flow sheet with a single CSTR unit.

    """
    if component_system is None:
        component_system = ComponentSystem(2)

    cstr = Cstr(component_system, "cstr")

    flow_sheet = FlowSheet(component_system)
    flow_sheet.add_unit(cstr)

    return flow_sheet


def setup_batch_elution_flow_sheet(component_system=None):
    """
    Set up a `FlowSheet` for a typical batch elution process.

    Parameters
    ----------
    component_system : ComponentSystem, optional
        The component system for the batch elution process.
        Defaults to a system with two components if None.

    Returns
    -------
    FlowSheet
        A flow sheet configured for batch elution processes, including feed and eluent
        inlets, a column, and an outlet.

    """
    if component_system is None:
        component_system = ComponentSystem(2)

    flow_sheet = FlowSheet(component_system)

    feed = Inlet(component_system, name="feed")
    eluent = Inlet(component_system, name="eluent")
    column = LumpedRateModelWithoutPores(component_system, name="column")
    outlet = Outlet(component_system, name="outlet")

    flow_sheet.add_unit(feed)
    flow_sheet.add_unit(eluent)
    flow_sheet.add_unit(column)
    flow_sheet.add_unit(outlet)

    flow_sheet.add_connection(feed, column)
    flow_sheet.add_connection(eluent, column)
    flow_sheet.add_connection(column, outlet)

    return flow_sheet


def setup_ssr_flow_sheet(component_system=None):
    """
    Set up a `FlowSheet` for a steady state recycling (SSR) process.

    Parameters
    ----------
    component_system : ComponentSystem, optional
        The component system for the SSR process.
        Defaults to a system with two components if None.

    Returns
    -------
    FlowSheet
        A flow sheet configured for SSR elution, including feed and eluent inlets, a
        mixer tank, a column, and an outlet.

    """
    if component_system is None:
        component_system = ComponentSystem(2)

    flow_sheet = FlowSheet(component_system)

    feed = Inlet(component_system, name="feed")
    eluent = Inlet(component_system, name="eluent")
    cstr = Cstr(component_system, name="cstr")
    column = LumpedRateModelWithoutPores(component_system, name="column")
    outlet = Outlet(component_system, name="outlet")

    flow_sheet.add_unit(feed)
    flow_sheet.add_unit(eluent)
    flow_sheet.add_unit(cstr)
    flow_sheet.add_unit(column)
    flow_sheet.add_unit(outlet)

    flow_sheet.add_connection(feed, cstr)
    flow_sheet.add_connection(cstr, column)
    flow_sheet.add_connection(eluent, column)
    flow_sheet.add_connection(column, cstr)
    flow_sheet.add_connection(column, outlet)

    flow_sheet.add_eluent_inlet(eluent)
    flow_sheet.add_feed_inlet(feed)
    flow_sheet.add_product_outlet(outlet)

    return flow_sheet


class TestFlowSheet(unittest.TestCase):
    """Test general functionatlity of `FlowSheet` class."""

    def setUp(self):
        self.component_system = ComponentSystem(2)

        # Single Cstr
        self.single_cstr_flow_sheet = setup_single_cstr_flow_sheet(self.component_system)

        # Batch
        self.batch_flow_sheet = setup_batch_elution_flow_sheet(self.component_system)

        # SSR
        self.ssr_flow_sheet = setup_ssr_flow_sheet(self.component_system)

    def test_unit_names(self):
        unit_names_expected = ["feed", "eluent", "cstr", "column", "outlet"]

        unit_names = list(self.ssr_flow_sheet.units_dict.keys())

        self.assertEqual(self.ssr_flow_sheet.unit_names, unit_names_expected)
        self.assertEqual(unit_names, unit_names_expected)

        # Connection already exists
        duplicate_unit_name = Inlet(self.component_system, "feed")
        with self.assertRaises(CADETProcessError):
            self.batch_flow_sheet.add_unit(duplicate_unit_name)

    def test_inlets(self):
        self.assertIn(self.ssr_flow_sheet.feed, self.ssr_flow_sheet.inlets)
        self.assertIn(self.ssr_flow_sheet.eluent, self.ssr_flow_sheet.inlets)

    def test_outlets(self):
        self.assertIn(self.ssr_flow_sheet.outlet, self.ssr_flow_sheet.outlets)

    def test_cstrs(self):
        self.assertIn(self.ssr_flow_sheet.cstr, self.ssr_flow_sheet.cstrs)

    def test_connections(self):
        feed = self.ssr_flow_sheet["feed"]
        eluent = self.ssr_flow_sheet["eluent"]
        cstr = self.ssr_flow_sheet["cstr"]
        column = self.ssr_flow_sheet["column"]
        outlet = self.ssr_flow_sheet["outlet"]
        expected_connections = {
            feed: {
                "origins": None,
                "destinations": {
                    None: {
                        cstr: [None],
                    },
                },
            },
            eluent: {
                "origins": None,
                "destinations": {
                    None: {
                        column: [None],
                    },
                },
            },
            cstr: {
                "origins": {
                    None: {
                        feed: [None],
                        column: [None],
                    },
                },
                "destinations": {
                    None: {
                        column: [None],
                    },
                },
            },
            column: {
                "origins": {
                    None: {
                        cstr: [None],
                        eluent: [None],
                    },
                },
                "destinations": {
                    None: {
                        cstr: [None],
                        outlet: [None],
                    },
                },
            },
            outlet: {
                "origins": {
                    None: {
                        column: [None],
                    },
                },
                "destinations": None,
            },
        }

        self.assertDictEqual(self.ssr_flow_sheet.connections, expected_connections)

        self.assertTrue(self.ssr_flow_sheet.connection_exists(feed, cstr))
        self.assertTrue(self.ssr_flow_sheet.connection_exists(eluent, column))
        self.assertTrue(self.ssr_flow_sheet.connection_exists(column, outlet))

        self.assertFalse(self.ssr_flow_sheet.connection_exists(feed, eluent))

    def test_name_decorator(self):
        feed = Inlet(self.component_system, name="feed")
        eluent = Inlet(self.component_system, name="eluent")
        cstr = Cstr(self.component_system, name="cstr")
        column = LumpedRateModelWithoutPores(self.component_system, name="column")
        outlet = Outlet(self.component_system, name="outlet")

        flow_sheet = FlowSheet(self.component_system)

        flow_sheet.add_unit(feed)
        flow_sheet.add_unit(eluent)
        flow_sheet.add_unit(cstr)
        flow_sheet.add_unit(column)
        flow_sheet.add_unit(outlet)

        flow_sheet.add_connection("feed", "cstr")
        flow_sheet.add_connection(cstr, column)
        flow_sheet.add_connection(eluent, "column")
        flow_sheet.add_connection(column, cstr)
        flow_sheet.add_connection("column", outlet)

        expected_connections = {
            feed: {
                "origins": None,
                "destinations": {
                    None: {
                        cstr: [None],
                    },
                },
            },
            eluent: {
                "origins": None,
                "destinations": {
                    None: {
                        column: [None],
                    },
                },
            },
            cstr: {
                "origins": {
                    None: {
                        feed: [None],
                        column: [None],
                    },
                },
                "destinations": {
                    None: {
                        column: [None],
                    },
                },
            },
            column: {
                "origins": {
                    None: {
                        cstr: [None],
                        eluent: [None],
                    },
                },
                "destinations": {
                    None: {
                        outlet: [None],
                        cstr: [None],
                    },
                },
            },
            outlet: {
                "origins": {
                    None: {
                        column: [None],
                    },
                },
                "destinations": None,
            },
        }
        self.assertDictEqual(flow_sheet.connections, expected_connections)

        # Connection already exists
        with self.assertRaises(CADETProcessError):
            flow_sheet.add_connection("column", "outlet")

        # Origin not found
        with self.assertRaises(CADETProcessError):
            flow_sheet.add_connection("wrong_origin", cstr)

        # Destination not found
        with self.assertRaises(CADETProcessError):
            flow_sheet.add_connection(cstr, "wrong_destination")

    def test_flow_rates(self):
        # Injection
        self.ssr_flow_sheet.feed.flow_rate = 0
        self.ssr_flow_sheet.eluent.flow_rate = 0
        self.ssr_flow_sheet.cstr.flow_rate = 1
        self.ssr_flow_sheet.set_output_state("column", 1)

        expected_flow_rates = {
            "feed": {
                "total_out": {
                    None: (0, 0, 0, 0),
                },
                "destinations": {
                    None: {
                        "cstr": {
                            None: (0, 0, 0, 0),
                        },
                    },
                },
            },
            "eluent": {
                "total_out": {
                    None: (0, 0, 0, 0),
                },
                "destinations": {
                    None: {
                        "column": {
                            None: (0, 0, 0, 0),
                        },
                    },
                },
            },
            "cstr": {
                "total_in": {
                    None: (0, 0, 0, 0),
                },
                "total_out": {
                    None: (1.0, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "feed": {
                            None: (0, 0, 0, 0),
                        },
                        "column": {
                            None: (0, 0, 0, 0),
                        },
                    },
                },
                "destinations": {
                    None: {
                        "column": {
                            None: (1.0, 0, 0, 0),
                        },
                    }
                },
            },
            "column": {
                "total_in": {
                    None: (1.0, 0, 0, 0),
                },
                "total_out": {
                    None: (1.0, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "cstr": {
                            None: (1.0, 0, 0, 0),
                        },
                        "eluent": {
                            None: (0, 0, 0, 0),
                        },
                    },
                },
                "destinations": {
                    None: {
                        "cstr": {
                            None: (0, 0, 0, 0),
                        },
                        "outlet": {
                            None: (1.0, 0, 0, 0),
                        },
                    },
                },
            },
            "outlet": {
                "origins": {
                    None: {
                        "column": {
                            None: (1.0, 0, 0, 0),
                        },
                    },
                },
                "total_in": {
                    None: (1.0, 0, 0, 0),
                },
            },
        }

        np.testing.assert_equal(
            self.ssr_flow_sheet.get_flow_rates(), expected_flow_rates
        )

        # Elution and Feed
        self.ssr_flow_sheet.feed.flow_rate = 1
        self.ssr_flow_sheet.eluent.flow_rate = 1
        self.ssr_flow_sheet.cstr.flow_rate = 0
        self.ssr_flow_sheet.set_output_state("column", 1)

        expected_flow_rates = {
            "feed": {
                "total_out": {
                    None: (1, 0, 0, 0),
                },
                "destinations": {
                    None: {
                        "cstr": {
                            None: (1, 0, 0, 0),
                        },
                    }
                },
            },
            "eluent": {
                "total_out": {None: (1, 0, 0, 0)},
                "destinations": {
                    None: {
                        "column": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
            },
            "cstr": {
                "total_in": {
                    None: (1.0, 0, 0, 0),
                },
                "total_out": {
                    None: (0.0, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "feed": {
                            None: (1, 0, 0, 0),
                        },
                        "column": {
                            None: (0, 0, 0, 0),
                        },
                    },
                },
                "destinations": {
                    None: {
                        "column": {
                            None: (0.0, 0, 0, 0),
                        },
                    },
                },
            },
            "column": {
                "total_in": {
                    None: (1.0, 0, 0, 0),
                },
                "total_out": {
                    None: (1.0, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "cstr": {
                            None: (0, 0, 0, 0),
                        },
                        "eluent": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
                "destinations": {
                    None: {
                        "cstr": {
                            None: (0, 0, 0, 0),
                        },
                        "outlet": {
                            None: (1.0, 0, 0, 0),
                        },
                    },
                },
            },
            "outlet": {
                "origins": {
                    None: {
                        "column": {
                            None: (1.0, 0, 0, 0),
                        },
                    },
                },
                "total_in": {
                    None: (1.0, 0, 0, 0),
                },
            },
        }
        np.testing.assert_equal(
            self.ssr_flow_sheet.get_flow_rates(), expected_flow_rates
        )

        # Elution
        self.ssr_flow_sheet.feed.flow_rate = 0
        self.ssr_flow_sheet.eluent.flow_rate = 1
        self.ssr_flow_sheet.cstr.flow_rate = 0
        self.ssr_flow_sheet.set_output_state("column", 1)

        expected_flow_rates = {
            "feed": {
                "total_out": {
                    None: (0, 0, 0, 0),
                },
                "destinations": {
                    None: {
                        "cstr": {
                            None: (0, 0, 0, 0),
                        },
                    },
                },
            },
            "eluent": {
                "total_out": {
                    None: (1, 0, 0, 0),
                },
                "destinations": {
                    None: {
                        "column": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
            },
            "cstr": {
                "total_in": {
                    None: (0, 0, 0, 0),
                },
                "total_out": {
                    None: (0, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "feed": {
                            None: (0, 0, 0, 0),
                        },
                        "column": {
                            None: (0, 0, 0, 0),
                        },
                    },
                },
                "destinations": {
                    None: {
                        "column": {
                            None: (0.0, 0, 0, 0),
                        },
                    },
                },
            },
            "column": {
                "total_in": {
                    None: (1.0, 0, 0, 0),
                },
                "total_out": {
                    None: (1.0, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "cstr": {
                            None: (0, 0, 0, 0),
                        },
                        "eluent": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
                "destinations": {
                    None: {
                        "cstr": {
                            None: (0, 0, 0, 0),
                        },
                        "outlet": {
                            None: (1.0, 0, 0, 0),
                        },
                    },
                },
            },
            "outlet": {
                "origins": {
                    None: {
                        "column": {
                            None: (1.0, 0, 0, 0),
                        },
                    },
                },
                "total_in": {
                    None: (1.0, 0, 0, 0),
                },
            },
        }

        np.testing.assert_equal(
            self.ssr_flow_sheet.get_flow_rates(), expected_flow_rates
        )

        # Recycle
        self.ssr_flow_sheet.feed.flow_rate = 0
        self.ssr_flow_sheet.eluent.flow_rate = 1
        self.ssr_flow_sheet.cstr.flow_rate = 0
        self.ssr_flow_sheet.set_output_state("column", 0)

        expected_flow_rates = {
            "feed": {
                "total_out": {
                    None: (0, 0, 0, 0),
                },
                "destinations": {
                    None: {
                        "cstr": {
                            None: (0, 0, 0, 0),
                        },
                    },
                },
            },
            "eluent": {
                "total_out": {
                    None: (1, 0, 0, 0),
                },
                "destinations": {
                    None: {
                        "column": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
            },
            "cstr": {
                "total_in": {
                    None: (1.0, 0, 0, 0),
                },
                "total_out": {
                    None: (0.0, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "feed": {
                            None: (0, 0, 0, 0),
                        },
                        "column": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
                "destinations": {
                    None: {
                        "column": {
                            None: (0.0, 0, 0, 0),
                        },
                    },
                },
            },
            "column": {
                "total_in": {
                    None: (1.0, 0, 0, 0),
                },
                "total_out": {
                    None: (1.0, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "cstr": {
                            None: (0, 0, 0, 0),
                        },
                        "eluent": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
                "destinations": {
                    None: {
                        "cstr": {
                            None: (1, 0, 0, 0),
                        },
                        "outlet": {
                            None: (0.0, 0, 0, 0),
                        },
                    },
                },
            },
            "outlet": {
                "origins": {
                    None: {
                        "column": {
                            None: (0, 0, 0, 0),
                        },
                    },
                },
                "total_in": {
                    None: (0.0, 0, 0, 0),
                },
            },
        }

        np.testing.assert_equal(
            self.ssr_flow_sheet.get_flow_rates(), expected_flow_rates
        )

        # Single Cstr
        expected_flow_rates = {
            "cstr": {
                "total_in": {
                    None: [0.0, 0.0, 0.0, 0.0],
                },
                "total_out": {None: [0.0, 0.0, 0.0, 0.0]},
                "origins": {},
                "destinations": {},
            }
        }

        np.testing.assert_equal(
            self.single_cstr_flow_sheet.get_flow_rates(), expected_flow_rates
        )

    def test_check_connectivity(self):
        self.assertTrue(self.single_cstr_flow_sheet.check_connections())
        self.assertTrue(self.batch_flow_sheet.check_connections())
        self.assertTrue(self.ssr_flow_sheet.check_connections())

        self.batch_flow_sheet.remove_unit("outlet")

        with self.assertWarns(Warning):
            self.batch_flow_sheet.check_connections()

    def test_output_state(self):
        column = self.ssr_flow_sheet.column

        output_state_expected = {None: [1, 0]}
        output_state = self.ssr_flow_sheet._output_states[column]
        np.testing.assert_equal(output_state, output_state_expected)

        output_state_expected = [1, 0]
        output_state = self.ssr_flow_sheet.output_states[column]
        np.testing.assert_equal(output_state, output_state_expected)

        self.ssr_flow_sheet.set_output_state(column, [0, 1])
        output_state_expected = {None: [0, 1]}
        output_state = self.ssr_flow_sheet._output_states[column]
        np.testing.assert_equal(output_state, output_state_expected)

        self.ssr_flow_sheet.set_output_state(column, 0)
        output_state_expected = {None: [1, 0]}
        output_state = self.ssr_flow_sheet._output_states[column]
        np.testing.assert_equal(output_state, output_state_expected)

        self.ssr_flow_sheet.set_output_state(column, [0.5, 0.5])
        output_state_expected = {None: [0.5, 0.5]}
        output_state = self.ssr_flow_sheet._output_states[column]
        np.testing.assert_equal(output_state, output_state_expected)

        self.ssr_flow_sheet.set_output_state(
            column,
            {
                "cstr": 0.1,
                "outlet": {
                    None: 0.9,
                },
            },
        )
        output_state_expected = {None: [0.1, 0.9]}
        output_state = self.ssr_flow_sheet._output_states[column]
        np.testing.assert_equal(output_state, output_state_expected)

        with self.assertRaises(TypeError):
            self.ssr_flow_sheet.set_output_state(column, "unknown_type")

        with self.assertRaises(CADETProcessError):
            self.ssr_flow_sheet.set_output_state(column, [1, 1])

        with self.assertRaises(CADETProcessError):
            self.ssr_flow_sheet.set_output_state(
                column,
                {
                    "column": 0.1,
                    "outlet": {
                        0: 0.9,
                    },
                },
            )

    def test_add_connection_error(self):
        """
        Test for all raised exceptions of add_connections.
        """
        inlet = self.ssr_flow_sheet["eluent"]
        column = self.ssr_flow_sheet["column"]
        outlet = self.ssr_flow_sheet["outlet"]
        external_unit = Cstr(self.component_system, name="external_unit")

        # Inlet can't be a destination
        with self.assertRaises(CADETProcessError):
            self.ssr_flow_sheet.add_connection(column, inlet)

        # Outlet can't be an origin
        with self.assertRaises(CADETProcessError):
            self.ssr_flow_sheet.add_connection(outlet, column)

        # Destination not part of flow_sheet
        with self.assertRaises(CADETProcessError):
            self.ssr_flow_sheet.add_connection(inlet, external_unit)

        # Origin not part of flow_sheet
        with self.assertRaises(CADETProcessError):
            self.ssr_flow_sheet.add_connection(external_unit, outlet)

        # Connection already exists
        with self.assertRaises(CADETProcessError):
            self.ssr_flow_sheet.add_connection(inlet, column)

        with self.assertRaisesRegex(CADETProcessError, "not a port of"):
            self.ssr_flow_sheet.set_output_state(column, [0.5, 0.5], "channel_5")


class TestCstrFlowRate(unittest.TestCase):
    """
    Test `Cstr` behaviour.

    Notes
    -----
    When the `flow_rate` parameter of the `Cstr` is not explicitly set, it is treated
    just like any other `UnitOperation`. I.e., q_in == q_out. In contrast, when a value
    is set, it has properties similar to an `Inlet`.
    """

    def setUp(self):
        self.component_system = ComponentSystem(2)

        flow_sheet = FlowSheet(self.component_system)

        inlet = Inlet(self.component_system, name="inlet")
        cstr = Cstr(self.component_system, name="cstr")
        outlet = Outlet(self.component_system, name="outlet")

        flow_sheet.add_unit(inlet)
        flow_sheet.add_unit(cstr)
        flow_sheet.add_unit(outlet)

        flow_sheet.add_connection(inlet, cstr)
        flow_sheet.add_connection(cstr, outlet)

        self.flow_sheet = flow_sheet

    def test_continuous_flow(self):
        self.flow_sheet.inlet.flow_rate = 1

        flow_rates = self.flow_sheet.get_flow_rates()

        cstr_in = flow_rates["cstr"]["total_in"][None]
        cstr_in_expected = [1.0, 0.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_in, cstr_in_expected)

        cstr_out = flow_rates["cstr"]["total_out"][None]
        cstr_out_expected = [1.0, 0.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_out, cstr_out_expected)

        # Test dynamic flow
        self.flow_sheet.inlet.flow_rate = [1, 1]

        flow_rates = self.flow_sheet.get_flow_rates()

        cstr_in = flow_rates["cstr"]["total_in"][None]
        cstr_in_expected = [1.0, 1.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_in, cstr_in_expected)

        cstr_out = flow_rates["cstr"]["total_out"][None]
        cstr_out_expected = [1.0, 1.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_out, cstr_out_expected)

    def test_no_flow(self):
        self.flow_sheet.inlet.flow_rate = 0

        flow_rates = self.flow_sheet.get_flow_rates()

        cstr_in = flow_rates["cstr"]["total_in"][None]
        cstr_in_expected = [0.0, 0.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_in, cstr_in_expected)

        cstr_out = flow_rates["cstr"]["total_out"][None]
        cstr_out_expected = [0.0, 0.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_out, cstr_out_expected)

        self.flow_sheet.inlet.flow_rate = 0
        self.flow_sheet.cstr.flow_rate = 0

        flow_rates = self.flow_sheet.get_flow_rates()

        cstr_in = flow_rates["cstr"]["total_in"][None]
        cstr_in_expected = [0.0, 0.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_in, cstr_in_expected)

        cstr_out = flow_rates["cstr"]["total_out"][None]
        cstr_out_expected = [0.0, 0.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_out, cstr_out_expected)

    def test_holdup(self):
        self.flow_sheet.inlet.flow_rate = 1
        self.flow_sheet.cstr.flow_rate = 0

        flow_rates = self.flow_sheet.get_flow_rates()

        cstr_in = flow_rates["cstr"]["total_in"][None]
        cstr_in_expected = [1.0, 0.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_in, cstr_in_expected)

        cstr_out = flow_rates["cstr"]["total_out"][None]
        cstr_out_expected = [0.0, 0.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_out, cstr_out_expected)

    def test_state_update(self):
        self.flow_sheet.inlet.flow_rate = [1, 1]

        state = {"flow_sheet.cstr.flow_rate": np.array([2, 2, 0, 0], ndmin=2)}

        flow_rates = self.flow_sheet.get_flow_rates(state)

        cstr_in = flow_rates["cstr"]["total_in"][None]
        cstr_in_expected = [1.0, 1.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_in, cstr_in_expected)

        cstr_out = flow_rates["cstr"]["total_out"][None]
        cstr_out_expected = [2.0, 2.0, 0.0, 0.0]
        np.testing.assert_almost_equal(cstr_out, cstr_out_expected)


class TestPorts(unittest.TestCase):
    def setUp(self):
        self.setup_mct_flow_sheet()

        self.setup_ccc_flow_sheet()

    def setup_mct_flow_sheet(self):
        self.component_system = ComponentSystem(1)

        mct_flow_sheet = FlowSheet(self.component_system)

        inlet = Inlet(self.component_system, name="inlet")
        mct_3c = MCT(self.component_system, nchannel=3, name="mct_3c")
        mct_2c1 = MCT(self.component_system, nchannel=2, name="mct_2c1")
        mct_2c2 = MCT(self.component_system, nchannel=2, name="mct_2c2")
        outlet1 = Outlet(self.component_system, name="outlet1")
        outlet2 = Outlet(self.component_system, name="outlet2")

        mct_flow_sheet.add_unit(inlet)
        mct_flow_sheet.add_unit(mct_3c)
        mct_flow_sheet.add_unit(mct_2c1)
        mct_flow_sheet.add_unit(mct_2c2)
        mct_flow_sheet.add_unit(outlet1)
        mct_flow_sheet.add_unit(outlet2)

        mct_flow_sheet.add_connection(inlet, mct_3c, destination_port="channel_0")
        mct_flow_sheet.add_connection(
            mct_3c, mct_2c1, origin_port="channel_0", destination_port="channel_0"
        )
        mct_flow_sheet.add_connection(
            mct_3c, mct_2c1, origin_port="channel_0", destination_port="channel_1"
        )
        mct_flow_sheet.add_connection(
            mct_3c, mct_2c2, origin_port="channel_1", destination_port="channel_0"
        )
        mct_flow_sheet.add_connection(mct_2c1, outlet1, origin_port="channel_0")
        mct_flow_sheet.add_connection(mct_2c1, outlet1, origin_port="channel_1")
        mct_flow_sheet.add_connection(mct_2c2, outlet2, origin_port="channel_0")

        self.mct_flow_sheet = mct_flow_sheet

    def setup_ccc_flow_sheet(self):
        self.component_system = ComponentSystem(1)

        ccc_flow_sheet = FlowSheet(self.component_system)

        inlet = Inlet(self.component_system, name="inlet")
        ccc1 = MCT(self.component_system, nchannel=2, name="ccc1")
        ccc2 = MCT(self.component_system, nchannel=2, name="ccc2")
        ccc3 = MCT(self.component_system, nchannel=2, name="ccc3")
        outlet = Outlet(self.component_system, name="outlet")

        ccc_flow_sheet.add_unit(inlet)
        ccc_flow_sheet.add_unit(ccc1)
        ccc_flow_sheet.add_unit(ccc2)
        ccc_flow_sheet.add_unit(ccc3)
        ccc_flow_sheet.add_unit(outlet)

        ccc_flow_sheet.add_connection(inlet, ccc1, destination_port="channel_0")
        ccc_flow_sheet.add_connection(
            ccc1, ccc2, origin_port="channel_0", destination_port="channel_0"
        )
        ccc_flow_sheet.add_connection(
            ccc2, ccc3, origin_port="channel_0", destination_port="channel_0"
        )
        ccc_flow_sheet.add_connection(ccc3, outlet, origin_port="channel_0")

        self.ccc_flow_sheet = ccc_flow_sheet

    def test_mct_connections(self):
        inlet = self.mct_flow_sheet["inlet"]
        mct_3c = self.mct_flow_sheet["mct_3c"]
        mct_2c1 = self.mct_flow_sheet["mct_2c1"]
        mct_2c2 = self.mct_flow_sheet["mct_2c2"]
        outlet1 = self.mct_flow_sheet["outlet1"]
        outlet2 = self.mct_flow_sheet["outlet2"]

        expected_connections = {
            inlet: {
                "origins": None,
                "destinations": {
                    None: {
                        mct_3c: ["channel_0"],
                    },
                },
            },
            mct_3c: {
                "origins": {
                    "channel_0": {
                        inlet: [None],
                    },
                    "channel_1": {},
                    "channel_2": {},
                },
                "destinations": {
                    "channel_0": {
                        mct_2c1: ["channel_0", "channel_1"],
                    },
                    "channel_1": {
                        mct_2c2: ["channel_0"],
                    },
                    "channel_2": {},
                },
            },
            mct_2c1: {
                "origins": {
                    "channel_0": {
                        mct_3c: ["channel_0"],
                    },
                    "channel_1": {
                        mct_3c: ["channel_0"],
                    },
                },
                "destinations": {
                    "channel_0": {
                        outlet1: [None],
                    },
                    "channel_1": {
                        outlet1: [None],
                    },
                },
            },
            mct_2c2: {
                "origins": {
                    "channel_0": {
                        mct_3c: ["channel_1"],
                    },
                    "channel_1": {},
                },
                "destinations": {
                    "channel_0": {
                        outlet2: [None],
                    },
                    "channel_1": {},
                },
            },
            outlet1: {
                "origins": {
                    None: {
                        mct_2c1: ["channel_0", "channel_1"],
                    },
                },
                "destinations": None,
            },
            outlet2: {
                "origins": {
                    None: {
                        mct_2c2: ["channel_0"],
                    },
                },
                "destinations": None,
            },
        }

        self.assertDictEqual(self.mct_flow_sheet.connections, expected_connections)

        self.assertTrue(
            self.mct_flow_sheet.connection_exists(inlet, mct_3c, destination_port="channel_0")
        )
        self.assertTrue(
            self.mct_flow_sheet.connection_exists(mct_3c, mct_2c1, "channel_0", "channel_0")
        )
        self.assertTrue(
            self.mct_flow_sheet.connection_exists(mct_3c, mct_2c1, "channel_0", "channel_1")
        )
        self.assertTrue(
            self.mct_flow_sheet.connection_exists(mct_3c, mct_2c2, "channel_1", "channel_0")
        )
        self.assertTrue(
            self.mct_flow_sheet.connection_exists(mct_2c1, outlet1, "channel_0")
        )
        self.assertTrue(
            self.mct_flow_sheet.connection_exists(mct_2c1, outlet1, "channel_1")
        )
        self.assertTrue(
            self.mct_flow_sheet.connection_exists(mct_2c2, outlet2, "channel_0")
        )

        self.assertFalse(
            self.mct_flow_sheet.connection_exists(mct_2c2, outlet2, "channel_1")
        )
        self.assertFalse(
            self.mct_flow_sheet.connection_exists(
                inlet, mct_2c1, destination_port="channel_0"
            )
        )

    def test_ccc_connections(self):
        inlet = self.ccc_flow_sheet["inlet"]
        ccc1 = self.ccc_flow_sheet["ccc1"]
        ccc2 = self.ccc_flow_sheet["ccc2"]
        ccc3 = self.ccc_flow_sheet["ccc3"]
        outlet = self.ccc_flow_sheet["outlet"]

        expected_connections = {
            inlet: {
                "origins": None,
                "destinations": {
                    None: {
                        ccc1: ["channel_0"],
                    },
                },
            },
            ccc1: {
                "origins": {
                    "channel_0": {
                        inlet: [None],
                    },
                    "channel_1": {},
                },
                "destinations": {
                    "channel_0": {
                        ccc2: ["channel_0"],
                    },
                    "channel_1": {},
                },
            },
            ccc2: {
                "origins": {
                    "channel_0": {
                        ccc1: ["channel_0"],
                    },
                    "channel_1": {},
                },
                "destinations": {
                    "channel_0": {
                        ccc3: ["channel_0"],
                    },
                    "channel_1": {},
                },
            },
            ccc3: {
                "origins": {
                    "channel_0": {
                        ccc2: ["channel_0"],
                    },
                    "channel_1": {},
                },
                "destinations": {
                    "channel_0": {
                        outlet: [None],
                    },
                    "channel_1": {},
                },
            },
            outlet: {
                "origins": {
                    None: {
                        ccc3: ["channel_0"],
                    },
                },
                "destinations": None,
            },
        }

        self.assertDictEqual(self.ccc_flow_sheet.connections, expected_connections)

        self.assertTrue(
            self.ccc_flow_sheet.connection_exists(
                inlet, ccc1, destination_port="channel_0"
            )
        )
        self.assertTrue(
            self.ccc_flow_sheet.connection_exists(ccc1, ccc2, "channel_0", "channel_0")
        )
        self.assertTrue(
            self.ccc_flow_sheet.connection_exists(ccc2, ccc3, "channel_0", "channel_0")
        )
        self.assertTrue(
            self.ccc_flow_sheet.connection_exists(ccc3, outlet, "channel_0")
        )

        self.assertFalse(self.ccc_flow_sheet.connection_exists(inlet, outlet))

    def test_mct_flow_rate_calculation(self):
        mct_flow_sheet = self.mct_flow_sheet

        mct_flow_sheet.inlet.flow_rate = 1

        inlet = self.mct_flow_sheet["inlet"]
        mct_3c = self.mct_flow_sheet["mct_3c"]
        mct_2c1 = self.mct_flow_sheet["mct_2c1"]
        mct_2c2 = self.mct_flow_sheet["mct_2c2"]

        #        mct_flow_sheet.set_output_state(mct_3c, [0.5, 0.5], 0)

        expected_flow = {
            "inlet": {
                "total_out": {None: (1, 0, 0, 0)},
                "destinations": {None: {"mct_3c": {"channel_0": (1, 0, 0, 0)}}},
            },
            "mct_3c": {
                "total_in": {
                    "channel_0": (1, 0, 0, 0),
                    "channel_1": (0, 0, 0, 0),
                    "channel_2": (0, 0, 0, 0),
                },
                "origins": {"channel_0": {"inlet": {None: (1, 0, 0, 0)}}},
                "total_out": {
                    "channel_0": (1, 0, 0, 0),
                    "channel_1": (0, 0, 0, 0),
                    "channel_2": (0, 0, 0, 0),
                },
                "destinations": {
                    "channel_0": {
                        "mct_2c1": {
                            "channel_0": (1, 0, 0, 0),
                            "channel_1": (0, 0, 0, 0),
                        }
                    },
                    "channel_1": {"mct_2c2": {"channel_0": (0, 0, 0, 0)}},
                },
            },
            "mct_2c1": {
                "total_in": {"channel_0": (1, 0, 0, 0), "channel_1": (0, 0, 0, 0)},
                "origins": {
                    "channel_0": {"mct_3c": {"channel_0": (1, 0, 0, 0)}},
                    "channel_1": {"mct_3c": {"channel_0": (0, 0, 0, 0)}},
                },
                "total_out": {"channel_0": (1, 0, 0, 0), "channel_1": (0, 0, 0, 0)},
                "destinations": {
                    "channel_0": {"outlet1": {None: (1, 0, 0, 0)}},
                    "channel_1": {"outlet1": {None: (0, 0, 0, 0)}},
                },
            },
            "mct_2c2": {
                "total_in": {"channel_0": (0, 0, 0, 0), "channel_1": (0, 0, 0, 0)},
                "origins": {
                    "channel_0": {"mct_3c": {"channel_1": (0, 0, 0, 0)}},
                },
                "total_out": {"channel_0": (0, 0, 0, 0), "channel_1": (0, 0, 0, 0)},
                "destinations": {"channel_0": {"outlet2": {None: (0, 0, 0, 0)}}},
            },
            "outlet1": {
                "total_in": {None: (1, 0, 0, 0)},
                "origins": {
                    None: {
                        "mct_2c1": {
                            "channel_0": (1, 0, 0, 0),
                            "channel_1": (0, 0, 0, 0),
                        }
                    }
                },
            },
            "outlet2": {
                "total_in": {None: (0, 0, 0, 0)},
                "origins": {None: {"mct_2c2": {"channel_0": (0, 0, 0, 0)}}},
            },
        }

        flow_rates = mct_flow_sheet.get_flow_rates()

        assert_almost_equal_dict(flow_rates, expected_flow)

    def test_port_add_connection(self):
        inlet = self.mct_flow_sheet["inlet"]
        mct_3c = self.mct_flow_sheet["mct_3c"]
        mct_2c2 = self.mct_flow_sheet["mct_2c2"]
        outlet1 = self.mct_flow_sheet["outlet1"]

        with self.assertRaises(CADETProcessError):
            self.mct_flow_sheet.add_connection(
                inlet, mct_3c, origin_port=0, destination_port=5
            )

        with self.assertRaises(CADETProcessError):
            self.mct_flow_sheet.add_connection(
                inlet, mct_3c, origin_port=5, destination_port=0
            )

        with self.assertRaises(CADETProcessError):
            self.mct_flow_sheet.add_connection(
                mct_2c2, outlet1, origin_port=0, destination_port=5
            )

    def test_set_output_state(self):
        mct_3c = self.mct_flow_sheet["mct_3c"]

        with self.assertRaises(CADETProcessError):
            self.mct_flow_sheet.set_output_state(mct_3c, [0.5, 0.5])

        with self.assertRaises(CADETProcessError):
            self.mct_flow_sheet.set_output_state(mct_3c, [0.5, 0.5], "channel_5")

        with self.assertRaises(CADETProcessError):
            self.mct_flow_sheet.set_output_state(
                mct_3c, {"mct_2c1": {"channel_0": 0.5, "channel_5": 0.5}}, "channel_0"
            )


class TestFlowRateMatrix(unittest.TestCase):
    """Test calculation of flow rates with another simple testcase by @daklauss"""

    def setUp(self):
        self.component_system = ComponentSystem(1)

        flow_sheet = FlowSheet(self.component_system)

        inlet = Inlet(self.component_system, name="inlet")
        cstr1 = Cstr(self.component_system, name="cstr1")
        cstr2 = Cstr(self.component_system, name="cstr2")
        outlet1 = Outlet(self.component_system, name="outlet1")
        outlet2 = Outlet(self.component_system, name="outlet2")

        flow_sheet.add_unit(inlet)
        flow_sheet.add_unit(cstr1)
        flow_sheet.add_unit(cstr2)

        flow_sheet.add_unit(outlet1)
        flow_sheet.add_unit(outlet2)

        flow_sheet.add_connection(inlet, cstr1)
        flow_sheet.add_connection(inlet, cstr2)

        flow_sheet.add_connection(cstr1, outlet1)
        flow_sheet.add_connection(cstr2, outlet2)

        flow_sheet.add_connection(cstr2, cstr1)

        flow_sheet.set_output_state(inlet, [0.3, 0.7])
        flow_sheet.set_output_state(cstr2, [0.5, 0.5])

        self.flow_sheet = flow_sheet

    def test_matrix_example(self):
        self.flow_sheet.inlet.flow_rate = [1]

        expected_flow_rates = {
            "inlet": {
                "total_out": {
                    None: (1, 0, 0, 0),
                },
                "destinations": {
                    None: {
                        "cstr1": {
                            None: (0.3, 0, 0, 0),
                        },
                        "cstr2": {
                            None: (0.7, 0, 0, 0),
                        },
                    },
                },
            },
            "cstr1": {
                "total_in": {
                    None: (0.65, 0, 0, 0),
                },
                "total_out": {
                    None: (0.65, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "inlet": {
                            None: (0.3, 0, 0, 0),
                        },
                        "cstr2": {
                            None: (0.35, 0, 0, 0),
                        },
                    }
                },
                "destinations": {
                    None: {
                        "outlet1": {
                            None: (0.65, 0, 0, 0),
                        },
                    },
                },
            },
            "cstr2": {
                "total_in": {
                    None: (0.7, 0, 0, 0),
                },
                "total_out": {
                    None: (0.7, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "inlet": {
                            None: (0.7, 0, 0, 0),
                        },
                    }
                },
                "destinations": {
                    None: {
                        "cstr1": {
                            None: (0.35, 0, 0, 0),
                        },
                        "outlet2": {
                            None: (0.35, 0, 0, 0),
                        },
                    },
                },
            },
            "outlet1": {
                "origins": {
                    None: {
                        "cstr1": {
                            None: (0.65, 0, 0, 0),
                        },
                    },
                },
                "total_in": {
                    None: (0.65, 0, 0, 0),
                },
            },
            "outlet2": {
                "origins": {
                    None: {
                        "cstr2": {
                            None: (0.35, 0, 0, 0),
                        },
                    }
                },
                "total_in": {
                    None: (0.35, 0, 0, 0),
                },
            },
        }

        calc_flow_rate = self.flow_sheet.get_flow_rates()

        assert_almost_equal_dict(calc_flow_rate, expected_flow_rates)


class TestFlowRateSelfMatrix(unittest.TestCase):
    """Test special case where one unit is connected to itself."""

    def setUp(self):
        self.component_system = ComponentSystem(1)

        flow_sheet = FlowSheet(self.component_system)

        inlet = Inlet(self.component_system, name="inlet")
        cstr = Cstr(self.component_system, name="cstr")
        outlet = Outlet(self.component_system, name="outlet")

        flow_sheet.add_unit(inlet)
        flow_sheet.add_unit(cstr)
        flow_sheet.add_unit(outlet)

        flow_sheet.add_connection(inlet, cstr)
        flow_sheet.add_connection(cstr, outlet)

        flow_sheet.add_connection(cstr, cstr)

        flow_sheet.set_output_state(cstr, [0.5, 0.5])

        self.flow_sheet = flow_sheet

    def test_matrix_self_example(self):
        self.flow_sheet.inlet.flow_rate = [1]

        expected_flow_rates = {
            "inlet": {
                "total_out": {
                    None: (1, 0, 0, 0),
                },
                "destinations": {
                    None: {
                        "cstr": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
            },
            "cstr": {
                "total_in": {
                    None: (2, 0, 0, 0),
                },
                "total_out": {
                    None: (2, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "inlet": {
                            None: (1, 0, 0, 0),
                        },
                        "cstr": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
                "destinations": {
                    None: {
                        "outlet": {
                            None: (1, 0, 0, 0),
                        },
                        "cstr": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
            },
            "outlet": {
                "total_in": {
                    None: (1, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "cstr": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
            },
        }
        calc_flow_rate = self.flow_sheet.get_flow_rates()
        np.testing.assert_equal(calc_flow_rate, expected_flow_rates)


class TestSingularFlowMatrix(unittest.TestCase):
    """Test cases with disconnected flow circles

    Notes
    -----
    `FlowSheet`: `Inlet` connected to an `Outlet` and two `Cstr`s connected to each
    other but not with the other units.

    """

    def setUp(self):
        self.component_system = ComponentSystem(1)

        flow_sheet = FlowSheet(self.component_system)

        inlet = Inlet(self.component_system, name="inlet")
        cstr1 = Cstr(self.component_system, name="cstr1")
        cstr2 = Cstr(self.component_system, name="cstr2")
        outlet = Outlet(self.component_system, name="outlet")

        flow_sheet.add_unit(inlet)
        flow_sheet.add_unit(cstr1)
        flow_sheet.add_unit(cstr2)
        flow_sheet.add_unit(outlet)

        flow_sheet.add_connection(inlet, outlet)

        flow_sheet.add_connection(cstr1, cstr2)
        flow_sheet.add_connection(cstr2, cstr1)

        self.flow_sheet = flow_sheet

    def test_expelled_cicuit(self):
        # Throw error even without flow, because system is singular
        flow_sheet = self.flow_sheet

        with self.assertRaises(CADETProcessError):
            flow_sheet.get_flow_rates()

        flow_sheet.inlet.flow_rate = [1]

        with self.assertRaises(CADETProcessError):
            flow_sheet.get_flow_rates()

    def test_expelled_circuit_with_flow(self):
        # Solvable because both disconnected circles have their own flow rates
        expected_flow_rates = {
            "inlet": {
                "total_out": {
                    None: (1, 0, 0, 0),
                },
                "destinations": {
                    None: {
                        "outlet": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
            },
            "cstr1": {
                "total_in": {
                    None: (1, 0, 0, 0),
                },
                "total_out": {
                    None: (1, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "cstr2": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
                "destinations": {
                    None: {
                        "cstr2": {None: (1, 0, 0, 0)},
                    },
                },
            },
            "cstr2": {
                "total_in": {
                    None: (1, 0, 0, 0),
                },
                "total_out": {
                    None: (1, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "cstr1": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
                "destinations": {
                    None: {
                        "cstr1": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
            },
            "outlet": {
                "total_in": {
                    None: (1, 0, 0, 0),
                },
                "origins": {
                    None: {
                        "inlet": {
                            None: (1, 0, 0, 0),
                        },
                    },
                },
            },
        }

        flow_sheet = self.flow_sheet
        flow_sheet.inlet.flow_rate = [1]
        flow_sheet.cstr1.flow_rate = [1]

        np.testing.assert_equal(flow_sheet.get_flow_rates(), expected_flow_rates)


if __name__ == "__main__":
    unittest.main()



import math
import warnings
from abc import abstractmethod
from typing import Any, Optional

import numpy as np

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import (
    Constant,
    NdPolynomial,
    Polynomial,
    SizedFloatList,
    SizedNdArray,
    SizedUnsignedList,
    String,
    Structure,
    Switch,
    UnsignedFloat,
    frozen_attributes,
)

from .binding import BindingBaseClass, NoBinding
from .componentSystem import ComponentSystem
from .discretization import (
    DiscretizationParametersBase,
    GRMDiscretizationDG,
    GRMDiscretizationFV,
    LRMDiscretizationDG,
    LRMDiscretizationFV,
    LRMPDiscretizationDG,
    LRMPDiscretizationFV,
    MCTDiscretizationFV,
    NoDiscretization,
)
from .reaction import BulkReactionBase, NoReaction, ParticleReactionBase
from .solutionRecorder import (
    CSTRRecorder,
    GRMRecorder,
    IORecorder,
    LRMPRecorder,
    LRMRecorder,
    MCTRecorder,
    TubularReactorRecorder,
)

__all__ = [
    "UnitBaseClass",
    "SourceMixin",
    "SinkMixin",
    "Inlet",
    "Outlet",
    "Cstr",
    "TubularReactorBase",
    "TubularReactor",
    "ChromatographicColumnBase",
    "LumpedRateModelWithoutPores",
    "LumpedRateModelWithPores",
    "GeneralRateModel",
    "MCT",
]


@frozen_attributes
class UnitBaseClass(Structure):
    """
    Base class for all UnitOperation classes.

    A UnitOperation object stores model parameters and states of a unit.
    Every unit operation can be assotiated with a binding behavior and a
    reaction model.
    UnitOperations can be connected in a FlowSheet.

    Attributes
    ----------
    n_comp : UnsignedInteger
        Number of components in a system.
    parameters : list
        list of parameter names.
    name : String
        name of the unit operation.
    has_ports : bool
        flag if unit has ports. The default is False.
    binding_model : BindingBaseClass
        binding behavior of the unit. Defaults to NoBinding.
    solution_recorder : IORecorder
        Solution recorder for the unit operation.

    See Also
    --------
    CADETProcess.processModel.binding
    CADETProcess.processModel.reaction
    CADETProcess.processModel.FlowSheet
    """

    name = String()

    _parameters = []
    _section_dependent_parameters = []
    _initial_state = []

    has_ports = False
    supports_binding = False
    supports_bulk_reaction = False
    supports_particle_reaction = False
    discretization_schemes = ()

    def __init__(
        self,
        component_system: ComponentSystem,
        name: str,
        *args: Optional[tuple],
        binding_model: Optional[BindingBaseClass] = None,
        bulk_reaction_model: Optional[BulkReactionBase] = None,
        particle_reaction_model: Optional[ParticleReactionBase] = None,
        discretization: Optional[DiscretizationParametersBase] = None,
        solution_recorder: Optional[IORecorder] = None,
        **kwargs: Optional[dict],
    ) -> None:
        """
        Initialize UnitBaseClass.

        Parameters
        ----------
        component_system : ComponentSystem
            The component system to which this unit belongs.
        name : str
            Name identifier for the unit instance.
        *args : tuple
            Positional arguments passed to the superclass constructor.
        binding_model : BindingBaseClass, optional
            A model describing specific binding interactions within the unit.
            If None, a default `NoBinding` model is used.
        bulk_reaction_model : BulkReactionBase, optional
            A model for bulk phase reactions. Defaults to `NoReaction` if not provided.
        particle_reaction_model : ParticleReactionBase, optional
            A model for reactions within particles. Defaults to `NoReaction`.
        discretization : DiscretizationParametersBase, optional
            Parameters defining spatial/temporal discretization of the unit.
            Defaults to `NoDiscretization` if None.
        solution_recorder : IORecorder, optional
            Object responsible for recording the simulation results.
            Defaults to a new `IORecorder` instance.
        **kwargs : dict
            Additional keyword arguments passed to the superclass constructor.
        """
        self.name = name
        self.component_system = component_system

        if binding_model is None:
            binding_model = NoBinding()
        self.binding_model = binding_model

        if bulk_reaction_model is None:
            bulk_reaction_model = NoReaction()
        self.bulk_reaction_model = bulk_reaction_model

        if particle_reaction_model is None:
            particle_reaction_model = NoReaction()
        self.particle_reaction_model = particle_reaction_model

        if discretization is None:
            discretization = NoDiscretization()
        self.discretization = discretization

        if solution_recorder is None:
            solution_recorder = IORecorder()
        self.solution_recorder = solution_recorder

        super().__init__(*args, **kwargs)

    @property
    def model(self) -> str:
        """str: Return model name."""
        return self.__class__.__name__

    @property
    def component_system(self) -> ComponentSystem:
        """ComponentSystem: Component of the unit operation."""
        return self._component_system

    @component_system.setter
    def component_system(self, component_system: ComponentSystem) -> None:
        if not isinstance(component_system, ComponentSystem):
            raise TypeError("Expected ComponentSystem")
        self._component_system = component_system

    @property
    def discretization(self) -> DiscretizationParametersBase:
        """DiscretizationParametersBase: Discretization parameters of the unit operation."""
        return self._discretization

    @discretization.setter
    def discretization(self, discretization: DiscretizationParametersBase) -> None:
        if not isinstance(discretization, DiscretizationParametersBase):
            raise TypeError("Expected DiscretizationParametersBase")

        if not isinstance(discretization, NoDiscretization):
            if not isinstance(discretization, self.discretization_schemes):
                raise CADETProcessError(
                    f"Unit does not support {type(discretization)}."
                )

        self._discretization = discretization

    @property
    def n_comp(self) -> int:
        """int: Number of components."""
        return self.component_system.n_comp

    @property
    def ports(self) -> list:
        """list: Ports of the unit operation."""
        return [None]

    @property
    def n_ports(self) -> int:
        """int: Number of ports."""
        return 1

    @property
    def parameters(self) -> dict:
        """dict: Dictionary with parameter values."""
        parameters = super().parameters

        if not isinstance(self.binding_model, NoBinding):
            parameters["binding_model"] = self.binding_model.parameters
        if not isinstance(self.bulk_reaction_model, NoReaction):
            parameters["bulk_reaction_model"] = self.bulk_reaction_model.parameters
        if not isinstance(self.particle_reaction_model, NoReaction):
            parameters['particle_reaction_model'] = \
                self.particle_reaction_model.parameters
        if not isinstance(self.discretization, NoDiscretization):
            parameters["discretization"] = self.discretization.parameters

        return parameters

    @parameters.setter
    def parameters(self, parameters: dict) -> None:
        try:
            self.binding_model.parameters = parameters.pop("binding_model")
        except KeyError:
            pass
        try:
            self.bulk_reaction_model.parameters = parameters.pop("bulk_reaction_model")
        except KeyError:
            pass
        try:
            self.particle_reaction_model.parameters = parameters.pop(
                "particle_reaction_model"
            )
        except KeyError:
            pass
        try:
            self.discretization.parameters = parameters.pop("discretization")
        except KeyError:
            pass

        super(UnitBaseClass, self.__class__).parameters.fset(self, parameters)

    @property
    def section_dependent_parameters(self) -> dict:
        """Return section dependent parameters."""
        parameters = {
            key: value
            for key, value in self.parameters.items()
            if key in self._section_dependent_parameters
        }
        return parameters

    @property
    def initial_state(self) -> dict:
        """dict: Dictionary with initial states."""
        initial_state = {st: getattr(self, st) for st in self._initial_state}

        return initial_state

    @initial_state.setter
    def initial_state(self, initial_state: dict) -> None:
        for st, value in initial_state.items():
            if st not in self._initial_state:
                raise CADETProcessError("Not a valid parameter")
            if value is not None:
                setattr(self, st, value)

    @property
    def missing_parameters(self) -> list:
        """Return list of missing parameters."""
        missing_parameters = super().missing_parameters

        missing_parameters += [
            f"binding_model.{param}" for param in self.binding_model.missing_parameters
        ]

        return missing_parameters

    def check_required_parameters(self) -> bool:
        """Checkf if there are missing parameters left."""
        if len(self.missing_parameters) == 0:
            return True
        else:
            for param in self.missing_parameters:
                warnings.warn(f'Unit {self.name}: Missing parameter "{param}".')
            return False

    @property
    def binding_model(self) -> BindingBaseClass:
        """
        BindingBaseClass: binding model parameters of the unit operation.

        Raises
        ------
        TypeError
            If binding_model object is not an instance of BindingBaseClass.
        CADETProcessError
            If number of components do not match.

        """
        return self._binding_model

    @binding_model.setter
    def binding_model(self, binding_model: BindingBaseClass) -> None:
        if not isinstance(binding_model, BindingBaseClass):
            raise TypeError("Expected BindingBaseClass")

        if not isinstance(binding_model, NoBinding):
            if not self.supports_binding:
                raise CADETProcessError("Unit does not support binding models.")

            if (
                binding_model.component_system is not self.component_system
                and not isinstance(binding_model, NoBinding)
            ):
                raise CADETProcessError("Component systems do not match.")

        self._binding_model = binding_model

    @property
    def n_bound_states(self) -> int:
        """int: Return number of bound states."""
        if isinstance(self.binding_model, NoBinding):
            return 0
        return self.binding_model.n_bound_states

    @property
    def bulk_reaction_model(self) -> BulkReactionBase:
        """
        BulkReactionBase: Reaction in bulk phase.

        Raises
        ------
        TypeError
            If bulk_reaction_model is not an instance of ReactionBaseClass.
        CADETProcessError
            If unit does not support bulk reaction model.
            If number of components do not match.

        """
        return self._bulk_reaction_model

    @bulk_reaction_model.setter
    def bulk_reaction_model(self, bulk_reaction_model: BulkReactionBase) -> None:
        if not isinstance(bulk_reaction_model, NoReaction):
            if not isinstance(bulk_reaction_model, BulkReactionBase):
                raise TypeError("Expected BulkReactionBase")

            if not self.supports_bulk_reaction:
                raise CADETProcessError("Unit does not support bulk reactions.")
            if bulk_reaction_model.component_system is not self.component_system:
                raise CADETProcessError("Component systems do not match.")

        self._bulk_reaction_model = bulk_reaction_model

    @property
    def particle_reaction_model(self) -> ParticleReactionBase:
        """
        ParticleReactionBase: Reaction in particle liquid phase.

        Raises
        ------
        TypeError
            If particle_reaction_model is not an instance of ReactionBaseClass.
        CADETProcessError
            If unit does not support particle reaction model.
            If number of components do not match.

        """
        return self._particle_reaction_model

    @particle_reaction_model.setter
    def particle_reaction_model(
        self, particle_reaction_model: ParticleReactionBase | BulkReactionBase
    ) -> None:
        if isinstance(particle_reaction_model, BulkReactionBase):
            try:
                warnings.warn(
                    "Detected Bulk Reaction Model. "
                    "Attempt casting to Particle Reaction Model."
                )
                particle_reaction_model = particle_reaction_model.to_particle_model()
            except NotImplementedError:
                pass

        if not isinstance(particle_reaction_model, NoReaction):
            if not isinstance(particle_reaction_model, ParticleReactionBase):
                raise TypeError("Expected ReactionBaseClass")

            if not self.supports_particle_reaction:
                raise CADETProcessError("Unit does not support particle reactions.")
            if particle_reaction_model.component_system is not self.component_system:
                raise CADETProcessError("Component systems do not match.")

        self._particle_reaction_model = particle_reaction_model

    def __repr__(self) -> str:
        """str: String-representation of the object."""
        return f"{self.__class__.__name__}(n_comp={self.n_comp}, name={self.name})"

    def __str__(self) -> str:
        """str: String-representation of the object."""
        return self.name


class SourceMixin(Structure):
    """
    Mixin class for Units that have Source-like behavior.

    See Also
    --------
    SinkMixin
    Inlet
    Cstr
    """

    _n_poly_coeffs = 4
    flow_rate = Polynomial(size=("_n_poly_coeffs"))
    _parameters = ["flow_rate"]
    _section_dependent_parameters = ["flow_rate"]


class SinkMixin:
    """
    Mixin class for Units that have Sink-like behavior.

    See Also
    --------
    SourceMixin
    Cstr
    """

    pass


class Inlet(UnitBaseClass, SourceMixin):
    """
    Pseudo unit operation model for streams entering the system.

    Attributes
    ----------
    c : NdPolynomial
        Polynomial coefficients for component concentration.
    flow_rate : NdPolynomial
        Polynomial coefficients for volumetric flow rate.
    solution_recorder : IORecorder
        Solution recorder for the unit operation.
    """

    c = NdPolynomial(size=("n_comp", "_n_poly_coeffs"), default=0)
    flow_rate = Polynomial(size=("_n_poly_coeffs"), default=0)
    _n_poly_coeffs = 4
    _parameters = ["c"]
    _section_dependent_parameters = (
        UnitBaseClass._section_dependent_parameters
        + SourceMixin._section_dependent_parameters
        + ["c"]
    )


class Outlet(UnitBaseClass, SinkMixin):
    """
    Pseudo unit operation model for streams leaving the system.

    Attributes
    ----------
    solution_recorder : IORecorder
        Solution recorder for the unit operation.
    """

    pass


class MixerSplitter(UnitBaseClass):
    """Pseudo unit operation for mixing/splitting streams in the system."""

    pass


class TubularReactorBase(UnitBaseClass):
    """
    Base class for tubular reactors and chromatographic columns.

    Provides methods for calculating geometric properties such as the cross
    section area and volume, as well as methods for convective and dispersive
    properties like mean residence time or NTP.

    Notes
    -----
        For subclassing, check that the total porosity and interstitial cross
        section area are computed correctly depending on the model porosities!

    Attributes
    ----------
    length : UnsignedFloat
        Length of column.
    diameter : UnsignedFloat
        Diameter of column.
    axial_dispersion : List of unsigned floats. Length depends on n_comp.
        Axial dispersion coefficient for each component.
    flow_direction : Switch
        If 1: Forward flow.
        If -1: Backwards flow.
    discretization : DiscretizationParametersBase
        Discretization scheme of the unit.
    """

    length = UnsignedFloat()
    diameter = UnsignedFloat()
    axial_dispersion = SizedUnsignedList(size="n_comp")
    flow_direction = Switch(valid=[-1, 1], default=1)
    c = SizedFloatList(size="n_comp", default=0)

    _initial_state = UnitBaseClass._initial_state + ["c"]
    _parameters = [
        "length",
        "diameter",
        "axial_dispersion",
        "flow_direction",
    ] + _initial_state
    _section_dependent_parameters = UnitBaseClass._section_dependent_parameters + [
        "axial_dispersion",
        "flow_direction",
    ]

    @property
    @abstractmethod
    def total_porosity(self) -> None:
        """Float | None: Total porosity of the unit operation."""
        pass

    @property
    def cross_section_area(self) -> float:
        """float: Cross section area of a Column.

        See Also
        --------
        volume
        cross_section_area_interstitial

        """
        if self.diameter is not None:
            return math.pi / 4 * self.diameter**2

    @cross_section_area.setter
    def cross_section_area(self, cross_section_area: float) -> None:
        self.diameter = (4 * cross_section_area / math.pi) ** 0.5

    def set_diameter_from_interstitial_velocity(self, Q: float, u0: float) -> None:
        """
        Set diamter from flow rate and interstitial velocity.

        In literature, often only the interstitial velocity is given.
        This method, the diameter / cross section area can be inferred from
        the flow rate, velocity, and porosity.


        Parameters
        ----------
        Q : float
            Volumetric flow rate.
        u0 : float
            Interstitial velocity.

        Notes
        -----
            Needs to be overwritten depending on the model porosities!
        """
        self.cross_section_area = Q / (u0 * self.total_porosity)

    @property
    def cross_section_area_interstitial(self) -> float:
        """float: Interstitial area between particles.

        Notes
        -----
            Needs to be overwritten depending on the model porosities!

        See Also
        --------
        cross_section_area

        """
        return self.total_porosity * self.cross_section_area

    @property
    def volume(self) -> float:
        """
        float: Volume of the TubularReactor.

        See Also
        --------
        cross_section_area

        """
        return self.cross_section_area * self.length

    @property
    def volume_interstitial(self) -> float:
        """
        float: Interstitial volume between particles.

        See Also
        --------
        cross_section_area

        """
        return self.cross_section_area_interstitial * self.length

    @property
    def volume_liquid(self) -> float:
        """float: Volume of the liquid phase."""
        return self.total_porosity * self.cross_section_area * self.length

    @property
    def volume_solid(self) -> float:
        """float: Volume of the solid phase."""
        return (1 - self.total_porosity) * self.cross_section_area * self.length

    def calculate_interstitial_rt(self, flow_rate: float) -> float:
        """
        Calculate mean residence time of a (non adsorbing) volume element.

        Parameters
        ----------
        flow_rate : float
            Volumetric flow rate.

        Returns
        -------
        t0 : float
            Mean residence time of packed bed.

        See Also
        --------
        calculate_interstitial_velocity
        calculate_superficial_rt
        """
        return self.volume_interstitial / flow_rate

    def calculate_superficial_rt(self, flow_rate: float) -> float:
        """
        Calculate mean residence time of a volume element in an empty column.

        Parameters
        ----------
        flow_rate : float
            Volumetric flow rate.

        Returns
        -------
        t_s : float
            Mean residence time of empty column.

        See Also
        --------
        calculate_superficial_velocity
        calculate_interstitial_rt
        """
        return self.volume / flow_rate

    def calculate_interstitial_velocity(self, flow_rate: float) -> float:
        """
        Calculate flow velocity of a (non adsorbing) volume element.

        Parameters
        ----------
        flow_rate : float
            Volumetric flow rate.

        Returns
        -------
        interstitial_velocity : float
            Interstitial flow velocity.

        See Also
        --------
        calculate_interstitial_rt
        calculate_superficial_velocity
        """
        return self.length / self.calculate_interstitial_rt(flow_rate)

    def calculate_superficial_velocity(self, flow_rate: float) -> float:
        """
        Calculate superficial flow velocity of a volume element in an empty column.

        Parameters
        ----------
        flow_rate : float
            Volumetric flow rate.

        Returns
        -------
        u_s : float
            Superficial flow velocity.

        See Also
        --------
        calculate_superficial_rt
        calculate_interstitial_velocity
        NTP
        """
        return self.length / self.calculate_superficial_rt(flow_rate)

    def calculate_flow_rate_from_velocity(self, u0: float) -> float:
        """
        Calculate volumetric flow rate from interstitial velocity.

        Parameters
        ----------
        u0 : float
            Interstitial flow velocity.

        Returns
        -------
        Q : float
            Volumetric flow rate.

        See Also
        --------
        calculate_interstitial_velocity
        calculate_interstitial_rt
        """
        return u0 * self.cross_section_area_interstitial

    def NTP(self, flow_rate: float) -> float:
        r"""
        Calculate number of theoretical plates.

        Parameters
        ----------
        flow_rate : float
            volumetric flow rate

        Calculated using the axial dispersion coefficient:

        .. math::
            NTP = \frac{u \cdot L_{Column}}{2 \cdot D_a}

        Returns
        -------
        NTP : float
            Number of theretical plates
        """
        u0 = self.calculate_interstitial_velocity(flow_rate)
        return u0 * self.length / (2 * np.array(self.axial_dispersion))

    def set_axial_dispersion_from_NTP(self, NTP: float, flow_rate: float) -> float:
        r"""
        Set axial dispersion from number of theoretical plates (NTP).

        Parameters
        ----------
        NTP : float
            Number of theroetical plates
        flow_rate : float
            volumetric flow rate

        Calculated using the axial dispersion coefficient:

        .. math::
            NTP = \frac{u_{int} \cdot L_{Column}}{2 \cdot D_{ax}}

        Returns
        -------
        NTP : float
            Number of theretical plates

        See Also
        --------
        calculate_interstitial_velocity
        NTP
        """
        u0 = self.calculate_interstitial_velocity(flow_rate)
        self.axial_dispersion = u0 * self.length / (2 * NTP)

    def calculate_bodenstein_number(self, flow_rate: float) -> float:
        r"""
        Calculate the Bodenstein number for a given flow rate.

        Parameters
        ----------
        flow_rate : float
            volumetric flow rate

        The Bodenstein number is calculated as follows:

        .. math::
            Bo = \frac{u_{int} \cdot L_{Column}}{D_{ax}},

        where $u_int$ is the interstitial velocity


        Returns
        -------
        Bo : float
            Bodenstein number

        See Also
        --------
        calculate_interstitial_velocity
        NTP
        """
        u0 = self.calculate_interstitial_velocity(flow_rate)
        return u0 * self.length / np.array(self.axial_dispersion)


class TubularReactor(TubularReactorBase):
    """
    Class for tubular reactors and tubing.

    Class can be used for a regular tubular reactor.

    Attributes
    ----------
    c : List of unsigned floats. Length depends on n_comp
        Initial concentration of the reactor.
    solution_recorder : TubularReactorRecorder
        Solution recorder for the unit operation.
    """

    supports_bulk_reaction = True
    discretization_schemes = (LRMDiscretizationFV, LRMDiscretizationDG)

    total_porosity = Constant(1)

    def __init__(
        self, *args: Any, discretization_scheme: Optional[str] = "FV", **kwargs: Any
    ) -> None:
        if discretization_scheme == "FV":
            discretization = LRMDiscretizationFV()
        elif discretization_scheme == "DG":
            discretization = LRMDiscretizationDG()

        super().__init__(
            *args,
            discretization=discretization,
            solution_recorder=TubularReactorRecorder(),
            **kwargs,
        )


class ChromatographicColumnBase(TubularReactorBase):
    """Base class for chromatogrpahic columns."""

    pass


class LumpedRateModelWithoutPores(ChromatographicColumnBase):
    """
    Parameters for a lumped rate model without pores.

    Attributes
    ----------
    total_porosity : UnsignedFloat between 0 and 1.
        Total porosity of the column.
    c : List of unsigned floats. Length depends on n_comp
            Initial concentration of the reactor.
    q : List of unsigned floats. Length depends on n_comp
        Initial concentration of the bound phase.
    solution_recorder : LRMRecorder
        Solution recorder for the unit operation.

    Notes
    -----
    Although technically the LumpedRateModelWithoutPores does not have
    particles, the particle reactions interface is used to support
    reactions in the solid phase and cross-phase reactions.
    """

    supports_binding = True
    supports_bulk_reaction = False
    supports_particle_reaction = True
    discretization_schemes = (LRMDiscretizationFV, LRMDiscretizationDG)

    total_porosity = UnsignedFloat(ub=1)
    _parameters = ["total_porosity"]

    _q = SizedUnsignedList(size="n_bound_states", default=0)

    _initial_state = ChromatographicColumnBase._initial_state + ["q"]
    _parameters = _parameters + _initial_state

    def __init__(
        self, *args: Any, discretization_scheme: Optional[str] = "FV", **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)

        if discretization_scheme == "FV":
            self.discretization = LRMDiscretizationFV()
        elif discretization_scheme == "DG":
            self.discretization = LRMDiscretizationDG()

        self.solution_recorder = LRMRecorder()

    @property
    def q(self) -> list:
        """list[float]: Initial solid phase concentration."""
        return self._q

    @q.setter
    def q(self, q: list) -> None:
        if isinstance(self.binding_model, NoBinding) and q not in (None, []):
            raise CADETProcessError("Cannot set q without binding model.")
        self._q = q

        self.parameters["q"] = self._q


class LumpedRateModelWithPores(ChromatographicColumnBase):
    """
    Parameters for the lumped rate model with pores.

    Attributes
    ----------
    bed_porosity : UnsignedFloat between 0 and 1.
        Porosity of the bed
    particle_porosity : UnsignedFloat between 0 and 1.
        Porosity of particles.
    particle_radius : UnsignedFloat
        Radius of the particles.
    film_diffusion : List of unsigned floats. Length depends on n_comp.
        Film diffusion coefficients for each component.
    pore_accessibility : List of unsigned floats. Length depends on n_comp.
        Accessibility of pores for components.
    c : List of unsigned floats. Length depends on n_comp
        Initial concentration of the reactor.
    cp : List of unsigned floats. Length depends on n_comp
        Initial concentration of the pores
    q : List of unsigned floats. Length depends on n_comp
        Initial concntration of the bound phase.
    solution_recorder : LRMPRecorder
        Solution recorder for the unit operation.
    """

    supports_binding = True
    supports_bulk_reaction = True
    supports_particle_reaction = True
    discretization_schemes = (LRMPDiscretizationFV, LRMPDiscretizationDG)

    bed_porosity = UnsignedFloat(ub=1)
    particle_porosity = UnsignedFloat(ub=1)
    particle_radius = UnsignedFloat()
    film_diffusion = SizedUnsignedList(size="n_comp")
    pore_accessibility = SizedUnsignedList(ub=1, size="n_comp", default=1)
    _parameters = [
        "bed_porosity",
        "particle_porosity",
        "particle_radius",
        "film_diffusion",
        "pore_accessibility",
    ]

    _section_dependent_parameters = TubularReactorBase._section_dependent_parameters + [
        "film_diffusion",
        "pore_accessibility",
    ]

    _cp = SizedUnsignedList(size="n_comp")
    _q = SizedUnsignedList(size="n_bound_states", default=0)

    _initial_state = ChromatographicColumnBase._initial_state + ["cp", "q"]
    _parameters = _parameters + _initial_state

    def __init__(
        self, *args: Any, discretization_scheme: Optional[str] = "FV", **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)

        if discretization_scheme == "FV":
            self.discretization = LRMPDiscretizationFV()
        elif discretization_scheme == "DG":
            self.discretization = LRMPDiscretizationDG()

        self.solution_recorder = LRMPRecorder()

    @property
    def total_porosity(self) -> float:
        """float: Total porosity of the column."""
        return self.bed_porosity + (1 - self.bed_porosity) * self.particle_porosity

    @property
    def cross_section_area_interstitial(self) -> float:
        """float: Interstitial area between particles.

        See Also
        --------
        cross_section_area
        """
        return self.bed_porosity * self.cross_section_area

    def set_diameter_from_interstitial_velocity(self, Q: float, u0: float) -> None:
        """
        Set diamter from flow rate and interstitial velocity.

        In literature, often only the interstitial velocity is given.
        This method, the diameter / cross section area can be inferred from
        the flow rate, velocity, and bed porosity.


        Parameters
        ----------
        Q : float
            Volumetric flow rate.
        u0 : float
            Interstitial velocity.

        Notes
        -----
            Overwrites parent method.
        """
        self.cross_section_area = Q / (u0 * self.bed_porosity)

    @property
    def cp(self) -> list[float]:
        """list[float]: Initial particle liquid concentration."""
        if self._cp is None:
            return self.c
        else:
            return self._cp

    @cp.setter
    def cp(self, cp: list[float]) -> None:
        self._cp = cp

        self.parameters["cp"] = cp

    @property
    def q(self) -> list[float]:
        """list[float]: Initial solid phase concentration."""
        return self._q

    @q.setter
    def q(self, q: list) -> None:
        if isinstance(self.binding_model, NoBinding) and q not in (None, []):
            raise CADETProcessError("Cannot set q without binding model.")
        self._q = q

        self.parameters["q"] = q


class GeneralRateModel(ChromatographicColumnBase):
    """
    Parameters for the general rate model.

    Attributes
    ----------
    bed_porosity : UnsignedFloat between 0 and 1.
        Porosity of the bed
    particle_porosity : UnsignedFloat between 0 and 1.
        Porosity of particles.
    particle_radius : UnsignedFloat
        Radius of the particles.
    film_diffusion : List of unsigned floats. Length depends on n_comp.
        Film diffusion coefficients for each component.
    pore_accessibility : List of unsigned floats. Length depends on n_comp.
        Accessibility of pores for components.
    pore_diffusion : List of unsigned floats. Length depends on n_comp.
        Diffusion rate for components in pore volume.
    surface_diffusion : List of unsigned floats. Length depends on n_comp.
        Diffusion rate for components in adsorbed state.
    c : List of unsigned floats. Length depends on n_comp
        Initial concentration in the mobile phase.
    cp : List of unsigned floats. Length depends on n_comp
        Initial concentration in the stagnant mobile phase within the pores.
    q : List of unsigned floats. Length depends on n_comp
        Initial concentration of the bound phase.
    solution_recorder : GRMRecorder
        Solution recorder for the unit operation.
    """

    supports_binding = True
    supports_bulk_reaction = True
    supports_particle_reaction = True
    discretization_schemes = (GRMDiscretizationFV, GRMDiscretizationDG)

    bed_porosity = UnsignedFloat(ub=1)
    particle_porosity = UnsignedFloat(ub=1)
    particle_radius = UnsignedFloat()
    film_diffusion = SizedUnsignedList(size="n_comp")
    pore_accessibility = SizedUnsignedList(ub=1, size="n_comp", default=1)
    pore_diffusion = SizedUnsignedList(size="n_comp")
    _surface_diffusion = SizedUnsignedList(size="n_bound_states")
    _parameters = [
        "bed_porosity",
        "particle_porosity",
        "particle_radius",
        "film_diffusion",
        "pore_accessibility",
        "pore_diffusion",
        "surface_diffusion",
    ]
    _section_dependent_parameters = TubularReactorBase._section_dependent_parameters + [
        "film_diffusion",
        "pore_accessibility",
        "pore_diffusion",
        "surface_diffusion",
    ]

    _cp = SizedUnsignedList(size="n_comp")
    _q = SizedUnsignedList(size="n_bound_states", default=0)

    _initial_state = ChromatographicColumnBase._initial_state + ["cp", "q"]
    _parameters = _parameters + _initial_state

    def __init__(
        self, *args: Any, discretization_scheme: Optional[str] = "FV", **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)

        if discretization_scheme == "FV":
            self.discretization = GRMDiscretizationFV()
        elif discretization_scheme == "DG":
            self.discretization = GRMDiscretizationDG()

        self.solution_recorder = GRMRecorder()

    @property
    def total_porosity(self) -> float:
        """float: Total porosity of the column."""
        return self.bed_porosity + (1 - self.bed_porosity) * self.particle_porosity

    @property
    def cross_section_area_interstitial(self) -> float:
        """float: Interstitial area between particles.

        See Also
        --------
        cross_section_area

        """
        return self.bed_porosity * self.cross_section_area

    def set_diameter_from_interstitial_velocity(self, Q: float, u0: float) -> None:
        """
        Set diamter from flow rate and interstitial velocity.

        In literature, often only the interstitial velocity is given.
        This method, the diameter / cross section area can be inferred from
        the flow rate, velocity, and bed porosity.


        Parameters
        ----------
        Q : float
            Volumetric flow rate.
        u0 : float
            Interstitial velocity.

        Notes
        -----
            Overwrites parent method.
        """
        self.cross_section_area = Q / (u0 * self.bed_porosity)

    @property
    def cp(self) -> list[float]:
        """list[float]: Initial particle liquid concentration."""
        if self._cp is None:
            return self.c
        else:
            return self._cp

    @cp.setter
    def cp(self, cp: list[float]) -> None:
        self._cp = cp

        self.parameters["cp"] = cp

    @property
    def q(self) -> list[float]:
        """list[float]: Initial solid phase concentration."""
        return self._q

    @q.setter
    def q(self, q: list[float]) -> None:
        if isinstance(self.binding_model, NoBinding) and q not in (None, []):
            raise CADETProcessError("Cannot set q without binding model.")
        self._q = q

        self.parameters["q"] = q

    @property
    def surface_diffusion(self) -> list[float]:
        """list[float] | None: Surface diffusion coefficients."""
        return self._surface_diffusion

    @surface_diffusion.setter
    def surface_diffusion(self, surface_diffusion: list[float]) -> None:
        if isinstance(self.binding_model, NoBinding):
            raise CADETProcessError(
                "Cannot set surface diffusion without binding model."
            )
        self._surface_diffusion = surface_diffusion

        self.parameters["surface_diffusion"] = surface_diffusion


class Cstr(UnitBaseClass, SourceMixin, SinkMixin):
    """
    Parameters for an ideal mixer.

    Parameters
    ----------
    c : List of unsigned floats. Length depends on n_comp
        Initial concentration of the reactor.
    q : List of unsigned floats. Length depends on n_comp
        Initial concentration of the bound phase.
    V : unsigned float
        Initial volume of the reactor.
    porosity : UnsignedFloat between 0 and 1.
        Total porosity of the Cstr.
    initial_liquid_volume : UnsignedFloat above 0.
        Initial liquid volume of the reactor.
    const_solid_volume : UnsignedFloat above or equal 0.
        Initial and constant solid volume of the reactor.
    flow_rate_filter: float
        Flow rate of pure liquid without components to reduce volume.
    solution_recorder : CSTRRecorder
        Solution recorder for the unit operation.

    Notes
    -----
    CADET generally supports particle reactions for the CSTR, however, this is currently
    not exposed since there are some issues with the interface (.
    """

    supports_binding = True
    supports_bulk_reaction = True
    supports_particle_reaction = False

    flow_rate_filter = UnsignedFloat(default=0)
    _parameters = ["const_solid_volume", "flow_rate_filter"]

    _section_dependent_parameters = (
        UnitBaseClass._section_dependent_parameters
        + SourceMixin._section_dependent_parameters
        + ["flow_rate_filter"]
    )

    c = SizedFloatList(size="n_comp", default=0)
    _q = SizedUnsignedList(size="n_bound_states", default=0)
    init_liquid_volume = UnsignedFloat()
    const_solid_volume = UnsignedFloat(default=0)
    _V = UnsignedFloat()
    _initial_state = UnitBaseClass._initial_state + ["c", "q", "init_liquid_volume"]
    _parameters = _parameters + _initial_state

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.solution_recorder = CSTRRecorder()

    @property
    def required_parameters(self) -> list:
        """
        Remove 'flow_rate' from required parameters.

        If flow rate is None, Q_in == Q_out'.
        """
        required_parameters = super().required_parameters.copy()
        required_parameters.remove("flow_rate")
        return required_parameters

    @property
    def porosity(self) -> float:
        """float: Porosity of the unit operation."""
        if self.const_solid_volume is None or self.init_liquid_volume is None:
            return None
        return self.init_liquid_volume / (
            self.init_liquid_volume + self.const_solid_volume
        )

    @porosity.setter
    def porosity(self, porosity: float) -> None:
        warnings.warn(
            "Field POROSITY is only supported for backwards compatibility, but the implementation "
            "of the CSTR has changed, please refer to the documentation. The POROSITY will be "
            "used to compute the constant solid volume from the total volume V."
        )
        if self.V is None:
            raise RuntimeError("Please set the volume first before setting a porosity.")
        self.const_solid_volume = self.V * (1 - porosity)
        self.init_liquid_volume = self.V * porosity

    @property
    def V(self) -> float:
        """float: Total volume of the unit operation."""
        return self._V

    @V.setter
    def V(self, V: float) -> None:
        warnings.warn(
            "The field V is only supported for backwards compatibility. "
            "Please set initial_liquid_volume and const_solid_volume."
        )
        self.init_liquid_volume = V
        self._V = V

    @property
    def volume(self) -> float:
        """float: Alias for volume."""
        return self.const_solid_volume + self.init_liquid_volume

    @property
    def volume_liquid(self) -> float:
        """float: Volume of the liquid phase."""
        return self.init_liquid_volume

    @property
    def volume_solid(self) -> float:
        """float: Volume of the solid phase."""
        return self.const_solid_volume

    def calculate_interstitial_rt(self, flow_rate: float) -> float:
        """
        Calculate mean residence time of a (non adsorbing) volume element.

        Parameters
        ----------
        flow_rate : float
            Volumetric flow rate.

        Returns
        -------
        t0 : float
            Mean residence time

        See Also
        --------
        calculate_interstitial_velocity
        """
        return self.volume_liquid / flow_rate

    @property
    def q(self) -> list:
        """list[float]: Initial solid phase concentration."""
        return self._q

    @q.setter
    def q(self, q: list) -> None:
        if isinstance(self.binding_model, NoBinding) and q not in (None, []):
            raise CADETProcessError("Cannot set q without binding model.")
        self._q = q

        self.parameters["q"] = q


class MCT(UnitBaseClass):
    """
    Parameters for multi-channel transportmodel.

    Parameters
    ----------
    length : UnsignedFloat
        Length of column.
    channel_cross_section_areas : List of unsinged floats. Lenght depends on nchannel.
        Diameter of column.
    axial_dispersion : List of unsigned floats. Length depends on n_comp and nchannel.
        Axial dispersion coefficient for component and each component.
    flow_direction : Switch
        If 1: Forward flow.
        If -1: Backwards flow.
    c : List of unsigned floats. Length depends n_comp or n_comp*nchannel.
        Initial concentration for components.
    exchange_matrix : List of unsigned floats. Lenght depends on nchannel.
    solution_recorder : MCTRecorder
        Solution recorder for the unit operation.
    n_channel : int
        Number of channels
    """

    has_ports = True
    supports_bulk_reaction = True

    discretization_schemes = MCTDiscretizationFV

    length = UnsignedFloat()
    channel_cross_section_areas = SizedFloatList(size="nchannel")
    axial_dispersion = SizedUnsignedList(size=("n_comp", "nchannel"))
    flow_direction = Switch(valid=[-1, 1], default=1)

    exchange_matrix = SizedNdArray(size=("nchannel", "nchannel", "n_comp"))

    _parameters = [
        "length",
        "channel_cross_section_areas",
        "axial_dispersion",
        "flow_direction",
        "exchange_matrix",
        "nchannel",
    ]
    _section_dependent_parameters = UnitBaseClass._section_dependent_parameters + [
        "axial_dispersion",
        "flow_direction",
    ]

    _section_dependent_parameters = UnitBaseClass._section_dependent_parameters + []

    c = SizedNdArray(size=("n_comp", "nchannel"), default=0)
    _initial_state = UnitBaseClass._initial_state + ["c"]
    _parameters = _parameters + _initial_state

    def __init__(self, *args: Any, nchannel: int, **kwargs: Any) -> None:
        """Initialize MCT."""
        discretization = MCTDiscretizationFV()
        self._nchannel = nchannel
        super().__init__(
            *args,
            discretization=discretization,
            solution_recorder=MCTRecorder(),
            **kwargs,
        )

    @property
    def nchannel(self) -> int:
        """int: The number of channels in the MCT."""
        return self._nchannel

    @nchannel.setter
    def nchannel(self, nchannel: int) -> None:
        self._nchannel = nchannel

    @property
    def ports(self) -> list:
        """list[str]: Ports of the unit operation."""
        return [f"channel_{i}" for i in range(self.nchannel)]

    @property
    def n_ports(self) -> int:
        """int: Number of ports (here the number of channels)."""
        return self.nchannel

    @property
    def volume(self) -> float:
        """
        float: Combined Volumes of all channels.

        See Also
        --------
        channel_cross_section_areas

        """
        return sum(self.channel_cross_section_areas) * self.length

    @property
    def volume_liquid(self) -> float:
        """float: Volume of the liquid phase. Equals the volume, since there is no solid phase."""
        return self.volume

    @property
    def volume_solid(self) -> float:
        """float: Volume of the solid phase. Equals zero, since there is no solid phase."""
        return 0
